from typing import Optional

from colorama import Fore, Style

CLOUD_CONSOLE_URL = "https://cloud.abstra.io"


class DeployMessages:
    @staticmethod
    def start(method: str = "upload"):
        method_label = "git push" if method == "git" else "upload"
        print(
            Fore.CYAN
            + Style.BRIGHT
            + f"\n[Deploy] Starting deploy via {method_label}...\n"
            + Style.RESET_ALL
        )

    @staticmethod
    def checking_linters():
        print(Fore.CYAN + "[Deploy] Checking for linter issues..." + Style.RESET_ALL)

    @staticmethod
    def authenticating():
        print(Fore.CYAN + "[Deploy] Verifying credentials..." + Style.RESET_ALL)

    @staticmethod
    def packaging():
        print(Fore.CYAN + "[Deploy] Packaging project files..." + Style.RESET_ALL)

    @staticmethod
    def uploading():
        print(
            Fore.CYAN + "[Deploy] Uploading files to Abstra Cloud..." + Style.RESET_ALL
        )

    @staticmethod
    def finalizing():
        print(Fore.CYAN + "[Deploy] Finalizing deploy..." + Style.RESET_ALL)

    @staticmethod
    def pushing(branch: str):
        print(
            Fore.CYAN
            + f"[Deploy] Pushing changes to branch '{branch}'..."
            + Style.RESET_ALL
        )

    @staticmethod
    def success(project_id: Optional[str] = None):
        if project_id:
            builds_url = f"{CLOUD_CONSOLE_URL}/projects/{project_id}/builds"
            print(
                Fore.GREEN
                + Style.BRIGHT
                + "\n[Deploy] Files uploaded successfully! "
                + Style.RESET_ALL
                + f"Track your build progress at: {Fore.GREEN}{builds_url}{Style.RESET_ALL}\n"
            )
        else:
            print(
                Fore.GREEN
                + Style.BRIGHT
                + "\n[Deploy] Files uploaded successfully! "
                + Style.RESET_ALL
                + f"Track your build progress at: {Fore.GREEN}{CLOUD_CONSOLE_URL}{Style.RESET_ALL}\n"
            )

    @staticmethod
    def error(message: str):
        print(
            Fore.RED + Style.BRIGHT + f"\n[Deploy] Error: {message}\n" + Style.RESET_ALL
        )

    @staticmethod
    def no_credentials():
        print(
            Fore.RED
            + Style.BRIGHT
            + "\n[Deploy] No credentials found. "
            + Style.RESET_ALL
            + "Please run 'abstra editor' and login via the web interface.\n"
        )



import unittest
from pathlib import Path, PurePosixPath, PureWindowsPath

from abstra_internals.services.file_watcher import FileWatcher


class TestFileWatcherShouldIgnorePath(unittest.TestCase):
    """Test suite for FileWatcher.should_ignore_path method."""

    def setUp(self):
        """Set up a FileWatcher instance for testing."""
        self.watcher = FileWatcher(handlers=[])

    def test_ignore_abstra_directory(self):
        """Test that .abstra/ directory is ignored."""
        self.assertTrue(self.watcher.should_ignore_path(Path("/project/.abstra/cache")))
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.abstra/settings.json"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/deep/path/.abstra/file.txt"))
        )

    def test_ignore_venv_directory(self):
        """Test that .venv directory is ignored."""
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.venv/lib/python"))
        )
        self.assertTrue(self.watcher.should_ignore_path(Path("/project/.venv")))
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/nested/path/.venv/bin/activate"))
        )

    def test_ignore_pycache_directory(self):
        """Test that __pycache__ directory is ignored."""
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/__pycache__/module.pyc"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/src/__pycache__"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/deep/nested/__pycache__/cache.pyc"))
        )

    def test_ignore_git_lock_files(self):
        """Test that specific .git lock files are ignored."""
        # Test index.lock
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/index.lock"))
        )

        # Test HEAD.lock
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/HEAD.lock"))
        )

        # Test config.lock (the original bug case)
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/config.lock"))
        )

    def test_ignore_git_directories(self):
        """Test that specific .git subdirectories are ignored."""
        # Test refs/
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/refs/heads/main"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/refs/tags/v1.0"))
        )

        # Test objects/
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/objects/ab/cdef123"))
        )

        # Test hooks/
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/hooks/pre-commit"))
        )

        # Test logs/
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/logs/HEAD"))
        )

    def test_ignore_git_special_files(self):
        """Test that specific .git files are ignored."""
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/COMMIT_EDITMSG"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/MERGE_HEAD"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/FETCH_HEAD"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/ORIG_HEAD"))
        )

    def test_ignore_any_git_lock_or_tmp_files(self):
        """Test that any .lock or .tmp file in .git directory is ignored."""
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/some-random.lock"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/temp-file.tmp"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(
                Path("/project/.git/refs/heads/branch.lock")
            )
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/objects/pack/temp.tmp"))
        )

    def test_do_not_ignore_regular_files(self):
        """Test that regular files are not ignored."""
        self.assertFalse(self.watcher.should_ignore_path(Path("/project/main.py")))
        self.assertFalse(self.watcher.should_ignore_path(Path("/project/src/utils.py")))
        self.assertFalse(self.watcher.should_ignore_path(Path("/project/README.md")))
        self.assertFalse(self.watcher.should_ignore_path(Path("/project/config.json")))

    def test_do_not_ignore_git_config_without_lock(self):
        """Test that .git/config (without .lock) is not ignored."""
        self.assertFalse(self.watcher.should_ignore_path(Path("/project/.git/config")))
        self.assertFalse(self.watcher.should_ignore_path(Path("/project/.git/HEAD")))

    def test_windows_paths_with_backslashes(self):
        """Test that Windows paths with backslashes are handled correctly."""
        # Create Windows-style paths using PureWindowsPath
        # These will have backslashes when converted to string
        win_path_1 = PureWindowsPath(r"C:\Users\project\.git\config.lock")
        win_path_2 = PureWindowsPath(
            r"C:\Abstra Seibel Sandbox\teste outro git\.git\config.lock"
        )
        win_path_3 = PureWindowsPath(r"D:\projects\myapp\.venv\lib\site-packages")
        win_path_4 = PureWindowsPath(r"C:\code\__pycache__\module.pyc")

        # These should all be ignored
        self.assertTrue(self.watcher.should_ignore_path(win_path_1))
        self.assertTrue(self.watcher.should_ignore_path(win_path_2))
        self.assertTrue(self.watcher.should_ignore_path(win_path_3))
        self.assertTrue(self.watcher.should_ignore_path(win_path_4))

    def test_windows_paths_with_spaces(self):
        """Test that Windows paths with spaces in directory names are handled correctly."""
        # The original bug case from the user
        win_path = PureWindowsPath(
            r"C:\Abstra Seibel Sandbox\teste outro git\.git\config.lock"
        )
        self.assertTrue(self.watcher.should_ignore_path(win_path))

        # More test cases with spaces
        self.assertTrue(
            self.watcher.should_ignore_path(
                PureWindowsPath(r"C:\Program Files\project\.git\index.lock")
            )
        )
        self.assertTrue(
            self.watcher.should_ignore_path(
                PureWindowsPath(r"D:\My Documents\code\.venv\lib")
            )
        )
        self.assertTrue(
            self.watcher.should_ignore_path(
                PureWindowsPath(r"C:\Users\John Doe\project\__pycache__")
            )
        )

    def test_unix_paths_with_forward_slashes(self):
        """Test that Unix paths with forward slashes work correctly."""
        self.assertTrue(
            self.watcher.should_ignore_path(
                PurePosixPath("/home/user/project/.git/config.lock")
            )
        )
        self.assertTrue(
            self.watcher.should_ignore_path(PurePosixPath("/var/www/app/.venv/bin"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(PurePosixPath("/opt/project/__pycache__"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(
                PurePosixPath("/Users/dev/project/.git/refs/heads/main")
            )
        )

    def test_mixed_case_sensitivity(self):
        """Test path matching with different cases."""
        # These should still match (patterns are case-sensitive)
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/__pycache__/file.pyc"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/config.lock"))
        )

        # These should not match (different case in pattern-sensitive parts)
        self.assertFalse(self.watcher.should_ignore_path(Path("/project/__PYCACHE__")))
        self.assertFalse(
            self.watcher.should_ignore_path(Path("/project/.GIT/config.lock"))
        )

    def test_edge_cases(self):
        """Test edge cases and boundary conditions."""
        # Root level paths
        self.assertTrue(self.watcher.should_ignore_path(Path("/.git/config.lock")))
        self.assertTrue(self.watcher.should_ignore_path(Path("/.venv")))

        # Very deep nested paths
        deep_path = Path("/a/b/c/d/e/f/g/.git/config.lock")
        self.assertTrue(self.watcher.should_ignore_path(deep_path))

        # Paths that contain pattern substring but shouldn't match
        self.assertFalse(self.watcher.should_ignore_path(Path("/project/.gitignore")))
        self.assertFalse(self.watcher.should_ignore_path(Path("/project/venv_backup")))
        self.assertFalse(self.watcher.should_ignore_path(Path("/project/cache_files")))

    def test_relative_paths(self):
        """Test that relative paths work correctly."""
        self.assertTrue(self.watcher.should_ignore_path(Path(".git/config.lock")))
        self.assertTrue(self.watcher.should_ignore_path(Path(".venv/lib/python")))
        self.assertTrue(
            self.watcher.should_ignore_path(Path("src/__pycache__/module.pyc"))
        )
        self.assertFalse(self.watcher.should_ignore_path(Path("src/main.py")))

    def test_git_subdirectories_ending_with_slash(self):
        """Test patterns that end with slash are matched correctly."""
        # Patterns like ".git/refs/" should match anything under refs
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/refs/heads/main"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/refs/tags/v1.0.0"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/objects/pack/file"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/hooks/pre-commit"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/logs/refs/heads/main"))
        )

    def test_abstra_directory_with_forward_slash(self):
        """Test that .abstra/ pattern (ending with /) matches correctly."""
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.abstra/settings"))
        )
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.abstra/cache/data.json"))
        )

    def test_lock_and_tmp_files_only_in_git(self):
        """Test that .lock and .tmp files are only ignored inside .git directory."""
        # Should be ignored (inside .git)
        self.assertTrue(
            self.watcher.should_ignore_path(Path("/project/.git/custom.lock"))
        )
        self.assertTrue(self.watcher.should_ignore_path(Path("/project/.git/temp.tmp")))

        # Should NOT be ignored (outside .git)
        self.assertFalse(
            self.watcher.should_ignore_path(Path("/project/database.lock"))
        )
        self.assertFalse(self.watcher.should_ignore_path(Path("/project/cache.tmp")))
        self.assertFalse(
            self.watcher.should_ignore_path(Path("/project/data/file.lock"))
        )

    def test_path_normalization_consistency(self):
        """Test that path normalization works consistently across different path representations."""
        # Unix-style path
        posix_path = PurePosixPath("/project/.git/config.lock")
        self.assertTrue(self.watcher.should_ignore_path(posix_path))

        # Windows-style path with backslashes
        win_path = PureWindowsPath(r"C:\project\.git\config.lock")
        self.assertTrue(self.watcher.should_ignore_path(win_path))

        # Windows-style path with forward slashes (also valid on Windows)
        win_path_forward = PureWindowsPath("C:/project/.git/config.lock")
        self.assertTrue(self.watcher.should_ignore_path(win_path_forward))


if __name__ == "__main__":
    unittest.main()



from os.path import join

from tests.fixtures import BaseTest


class TestCRUDFiles(BaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.client = self.get_editor_flask_client()

    def test_api_update(self):
        self.client.put(
            "/_editor/api/workspace",
            json={"name": "test-workspace-updated", "brand_name": "test-brand-name"},
        )
        workspace = self.client.get("/_editor/api/workspace").get_json()
        self.assertEqual(workspace["name"], "test-workspace-updated")
        self.assertEqual(workspace["brand_name"], "test-brand-name")

    def test_list_files(self):
        res = self.client.get("/_editor/api/workspace/files")
        self.assertEqual(res.status_code, 200)

        self.assertCountEqual(
            res.json or [],
            [
                dict(path="requirements.txt", name="requirements.txt", type="file"),
                dict(path="abstra.json", name="abstra.json", type="file"),
                dict(name="favicon.ico", path="favicon.ico", type="file"),
                dict(name="logo.png", path="logo.png", type="file"),
                dict(name=".gitignore", path=".gitignore", type="file"),
            ],
        )

    def test_list_files_with_subdir(self):
        (self.root / "subdir").mkdir()
        (self.root / "subdir" / "test.txt").touch()

        res = self.client.get("/_editor/api/workspace/files?path=subdir")
        self.assertEqual(
            res.json,
            [dict(path=join("subdir", "test.txt"), name="test.txt", type="file")],
        )

    def test_shallow_module_mode(self):
        (self.root / "subdir").mkdir()
        (self.root / "subdir" / "__init__.py").touch()
        (self.root / "subdir" / "test.py").touch()

        res = self.client.get("/_editor/api/workspace/files?mode=module")
        if res.json is None:
            raise Exception("No json response")

        self.assertIn(dict(path="subdir", name="subdir", type="package"), res.json)

    def test_nested_module_mode(self):
        (self.root / "subdir").mkdir()
        (self.root / "subdir" / "__init__.py").touch()
        (self.root / "subdir" / "test.py").touch()

        res = self.client.get("/_editor/api/workspace/files?mode=module&path=subdir")
        if res.json is None:
            raise Exception("No json response")

        self.assertIn(dict(path="subdir.test", name="test", type="module"), res.json)



import io
from typing import Optional, Union

from abstra_internals.entities.forms.widgets.file_upload import upload_widget_file
from abstra_internals.entities.forms.widgets.widget_base import OutputWidget


class FileOutput(OutputWidget):
    """File output widget for displaying downloadable files.

    Attributes:
        file (str): The URL of the file to make available for download.
    """

    type = "file-output"

    def __init__(
        self,
        file: Union[str, io.IOBase],
        *,
        download_text: str = "Download",
        full_width: bool = False,
        filename_override: Optional[str] = None,
    ):
        """Initialize a FileOutput widget.

        Args:
            file (Union[str, io.IOBase]): The file to make available for download, as a URL or file-like object.
            download_text (str): Text displayed on the download button.
            full_width (bool): Whether the download button should take up the full width of its container.
            filename_override (str): Optional filename override to use for the download.
        """
        self.file = upload_widget_file(file, filename_override)
        self.download_text = download_text
        self.full_width = full_width

    def _render(self):
        return {
            "type": self.type,
            "fileUrl": self.file,
            "downloadText": self.download_text,
            "fullWidth": self.full_width,
        }



"""
Abstra Browser AI - ReAct Framework for AI-Powered Web Browsing

This package provides a ReAct (Reasoning + Acting) framework for AI-powered web browsing
using Selenium WebDriver and Abstra AI services.
"""

from .react_browser import experimental_prompt_browser

__all__ = ["experimental_prompt_browser"]



from unittest import TestCase

from abstra_internals.consts.filepaths import GITIGNORE_FILEPATH
from abstra_internals.repositories.linter.rules.env_in_bundle import EnvInBundle
from tests.fixtures import clear_dir, init_dir


class EnvInBundleTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_env_on_bundle_valid_default(self):
        rule = EnvInBundle()
        env_file = self.root / ".env"
        self.assertFalse(env_file.exists())
        self.assertEqual(len(rule.find_issues()), 0)

    def test_env_on_bundle_valid_with_env(self):
        rule = EnvInBundle()
        env_file = self.root / ".env"
        abstraignore_file = self.root / GITIGNORE_FILEPATH
        abstraignore_file.write_text(".env")
        env_file.touch()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_env_on_bundle_invalid_without_gitignore_file(self):
        env_file = self.root / ".env"
        env_file.touch()
        rule = EnvInBundle()
        self.assertEqual(len(rule.find_issues()), 1)

    def test_env_on_bundle_invalid_with_gitignore_file(self):
        env_file = self.root / ".env"
        env_file.touch()
        abstraignore_file = self.root / GITIGNORE_FILEPATH
        abstraignore_file.touch()
        rule = EnvInBundle()
        self.assertEqual(len(rule.find_issues()), 1)

    def test_env_on_bundle_fix(self):
        env_file = self.root / ".env"
        env_file.touch()
        rule = EnvInBundle()
        self.assertEqual(len(rule.find_issues()), 1)
        rule.find_issues()[0].fixes[0].fix()
        self.assertEqual(len(rule.find_issues()), 0)
        abstraignore_file = self.root / GITIGNORE_FILEPATH
        self.assertTrue(abstraignore_file.exists())
        with abstraignore_file.open("r") as file:
            content = file.read()
            self.assertTrue(".env" in content)



class CaseInsensitiveDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def og_key(self, key):
        for k in super().keys():
            if k.lower() == key.lower():
                return k

    def __setitem__(self, key, value):
        if self.og_key(key):
            super().__setitem__(self.og_key(key), value)
        else:
            super().__setitem__(key, value)

    def __getitem__(self, key):
        return super().__getitem__(self.og_key(key))

    def __delitem__(self, key):
        return super().__delitem__(self.og_key(key))

    def __contains__(self, key):
        return self.og_key(key) is not None

    def get(self, key, default=None):
        return super().get(self.og_key(key), default)



from typing import Dict, List, Optional, Union

import flask
from pydantic import Field

from abstra_internals.repositories.tasks import TaskDTO
from abstra_internals.utils.dict import filter_non_string_values
from abstra_internals.utils.serializable import Serializable


class Request(Serializable):
    query_params: Dict[str, str]
    headers: Dict[str, str]
    method: str
    body: str


class Response(Serializable):
    headers: Dict[str, str]
    status: int
    body: str


class ExecutionMock(Serializable):
    test_pending_tasks: List[TaskDTO] = Field(default_factory=list)


class HookExecutionMock(ExecutionMock):
    test_request: Optional[Request] = None


class FormExecutionMock(ExecutionMock):
    test_answers: List[Union[str, None]] = Field(default_factory=list)


class ScriptExecutionMock(ExecutionMock):
    test_trigger_task: Optional[TaskDTO] = None


class JobExecutionMock(ExecutionMock):
    pass


class CodeSnippetExecutionMock(ExecutionMock):
    pass


class HookContext(Serializable):
    request: Request
    response: Response
    sent_tasks: List[str] = Field(default_factory=list)
    legacy_thread_data: dict = Field(default_factory=dict)
    mock_execution: HookExecutionMock = Field(default_factory=HookExecutionMock)


class FormContext(Serializable):
    request: Request
    sent_tasks: List[str] = Field(default_factory=list)
    legacy_thread_data: dict = Field(default_factory=dict)
    mock_execution: FormExecutionMock = Field(default_factory=FormExecutionMock)


class ScriptContext(Serializable):
    task_id: str
    sent_tasks: List[str] = Field(default_factory=list)
    legacy_thread_data: dict = Field(default_factory=dict)
    mock_execution: ScriptExecutionMock = Field(default_factory=ScriptExecutionMock)


class JobContext(Serializable):
    sent_tasks: List[str] = Field(default_factory=list)
    legacy_thread_data: dict = Field(default_factory=dict)
    mock_execution: JobExecutionMock = Field(default_factory=JobExecutionMock)


class CodeSnippetContext(Serializable):
    mock_execution: CodeSnippetExecutionMock = Field(
        default_factory=CodeSnippetExecutionMock
    )


ClientContext = Union[
    HookContext, FormContext, ScriptContext, JobContext, CodeSnippetContext
]


def extract_flask_request(request: flask.Request) -> Request:
    return Request(
        headers=filter_non_string_values(dict(request.headers)),
        body=request.get_data(as_text=True),
        query_params={**request.args},
        method=request.method,
    )



from typing import Dict, List, Optional

from abstra_internals.entities.forms.form_state import State
from abstra_internals.entities.forms.template import (
    SPREADED_KEY,
    Template,
    TemplateRenderer,
)
from abstra_internals.entities.forms.widgets.widget_base import InputWidget


class Row(InputWidget):
    """
    Creates a horizontal layout for organizing other widgets side-by-side.

    Attributes:
        value (Dict[str, State]): A dictionary containing the current state/values of the nested widgets.
    """

    type = "row"
    widgets: Template
    value: State
    key: Optional[str]

    def __init__(
        self, widgets: "Template", *, key: Optional[str] = None, errors: List[str] = []
    ):
        """
        Initializes a Row widget.

        Args:
            widgets (Template): A list of widgets to display horizontally in this row.
            key (Optional[str]): An optional key to identify this row. If not provided, one will be generated.
            errors (List[str]): A list of error messages to start with. Defaults to [].
        """
        self.widgets = widgets
        self.value = State()
        self._errors = errors
        for idx, w in enumerate(self.widgets):
            if isinstance(w, InputWidget):
                key = w.key or w.type + str(idx)
                self.value[w.key or f"{self.key}{key}"] = w.value

        self.key = self.compute_key()
        self.required = any(
            w.required for w in self.widgets if isinstance(w, InputWidget)
        )

    def compute_key(self) -> str:
        return (
            SPREADED_KEY
            + "_row_"
            + "_".join(
                w.key if isinstance(w, InputWidget) and w.key else w.type
                for w in self.widgets
            )
        )

    def update_required(self):
        return any(w.required for w in self.widgets if isinstance(w, InputWidget))

    def _render(self):
        renderer = TemplateRenderer(self.widgets)
        output = renderer.render(self.value)
        rendered_widgets = output["widgets"]

        serialized_value = {
            widget["key"]: widget["value"]
            for widget in output["widgets"]
            if "key" in widget
        }

        errors = []
        for rendered_widget in rendered_widgets:
            if "errors" in rendered_widget and rendered_widget["errors"]:
                errors.append(
                    (rendered_widget["label"] or rendered_widget["key"]) + ":"
                )
                errors.extend(rendered_widget["errors"])

        return {
            "type": self.type,
            "key": self.key,
            "widgets": rendered_widgets,
            "value": serialized_value,
            "errors": errors,
        }

    def _parse_value(self, value: dict) -> Dict[str, State]:
        renderer = TemplateRenderer(self.widgets)

        keys = list(key for key in value.keys() if key.startswith(SPREADED_KEY))
        for key in keys:
            if key.startswith(SPREADED_KEY) and isinstance(value[key], dict):
                value.update(value[key])
        parsed = renderer.parse_state(raw_state=value, include_missing=True)
        return parsed



import json
from multiprocessing import Pipe
from pathlib import Path
from uuid import uuid4

from abstra_internals.controllers.execution.execution import ExecutionController
from abstra_internals.controllers.execution.execution_client_hook import HookClient
from abstra_internals.entities.execution_context import HookContext, Request, Response
from abstra_internals.repositories.project.project import HookStage
from tests.fixtures import BaseTest


class ExecutionControllerTest(BaseTest):
    def setUp(self) -> None:
        super().setUp()

        self.request = Request(
            body={"a": 1}.__str__(),
            headers={"auth": "some_secret_token"},
            query_params={"c": "3"},
            method="GET",
        )

        self.context = HookContext(
            request=Request(
                body={"a": 1}.__str__(),
                headers={"auth": "some_secret_token"},
                query_params={"c": "3"},
                method="GET",
            ),
            response=Response(headers={}, status=200, body=""),
        )

        self.project = self.repositories.project.load(include_disabled_stages=False)
        self.stage = HookStage.create(
            title="mock_stage",
            file="mock_file.py",
            id="mock_hook_id",
            workflow_position=(0, 1),
        )
        Path(self.stage.file).write_text("print('Hello, World!')", encoding="utf-8")
        self.project.add_stage(self.stage)
        self.repositories.project.save(self.project)

        self.parent_conn, child_conn = Pipe()
        self.hook_client = HookClient(
            self.context, conn=child_conn, production_mode=False
        )

    def test_run_initial_returns_dto(self):
        ExecutionController(
            repositories=self.repositories,
            stage=self.stage,
            context=self.context,
            client=self.hook_client,
        ).run(execution_id=uuid4().__str__(), worker_id="mock-worker-id")

        started_msg_str = self.parent_conn.recv()
        assert isinstance(started_msg_str, str), (
            f"Expected str, got {type(started_msg_str)}"
        )

        started_msg = json.loads(started_msg_str)
        assert started_msg["type"] == "execution:started"
        assert "executionId" in started_msg

        response = self.parent_conn.recv()

        if not response:
            self.fail("Response was not set")

        if not isinstance(response, Response):
            response = Response(
                headers=response.get("headers", {}),
                status=response.get("status", 200),
                body=response.get("body", ""),
            )

        self.assertEqual(response.status, 200)

    def test_execution_started_message_is_json_serializable(self):
        """Test that ExecutionStartedMessage is properly serialized to JSON"""
        ExecutionController(
            repositories=self.repositories,
            stage=self.stage,
            context=self.context,
            client=self.hook_client,
        ).run(execution_id="test_execution_id", worker_id="mock-worker-id")

        # Receive and verify the message is a valid JSON string
        started_msg_str = self.parent_conn.recv()
        self.assertIsInstance(started_msg_str, str)

        # Verify it can be parsed as JSON
        try:
            started_msg = json.loads(started_msg_str)
        except json.JSONDecodeError:
            self.fail("ExecutionStartedMessage is not valid JSON")

        # Verify the structure
        self.assertEqual(started_msg["type"], "execution:started")
        self.assertIn("executionId", started_msg)
        self.assertIsInstance(started_msg["executionId"], str)

    def test_execution_ended_message_is_json_serializable(self):
        """Test that ExecutionEndedMessage is properly serialized to JSON"""
        ExecutionController(
            repositories=self.repositories,
            stage=self.stage,
            context=self.context,
            client=self.hook_client,
        ).run(execution_id="test_execution_id", worker_id="mock-worker-id")

        # Skip the started message
        self.parent_conn.recv()

        # Skip response
        self.parent_conn.recv()

        # Get the ended message
        ended_msg_str = self.parent_conn.recv()
        self.assertIsInstance(ended_msg_str, str)

        # Verify it can be parsed as JSON
        try:
            ended_msg = json.loads(ended_msg_str)
        except json.JSONDecodeError:
            self.fail("ExecutionEndedMessage is not valid JSON")

        # Verify the structure
        self.assertEqual(ended_msg["type"], "execution:ended")
        self.assertIn("exitStatus", ended_msg)
        self.assertEqual(ended_msg["exitStatus"], "SUCCESS")



import time
from dataclasses import dataclass, field
from enum import Enum
from multiprocessing import Queue
from multiprocessing.process import BaseProcess
from threading import RLock, Thread
from typing import Any, Dict, List, Optional
from uuid import uuid4

from abstra_internals.controllers.execution.executor_config import (
    ExecutorConfig,
)
from abstra_internals.controllers.execution.executor_metrics import MetricsCollector
from abstra_internals.controllers.execution.executor_process import (
    ExecuteRequest,
    ExecutorCommand,
    ExecutorResponse,
    RabbitMQParams,
    ShutdownRequest,
    WarmupRequest,
    executor_main,
)
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.multiprocessing import MPContext
from abstra_internals.repositories.project.project import StageType, StageWithFile
from abstra_internals.utils.multiprocessing import safe_multiprocessing_queue


class ExecutorStatus(str, Enum):
    SPAWNING = "spawning"
    WARMING_UP = "warming_up"
    IDLE = "idle"
    BUSY = "busy"
    DEAD = "dead"


@dataclass
class ExecutorHandle:
    executor_id: str
    process: BaseProcess
    work_queue: Queue
    response_queue: Queue
    status: ExecutorStatus
    pid: Optional[int] = None
    is_form_reserved: bool = True
    executions_completed: int = 0
    current_execution_id: Optional[str] = None
    spawn_time: float = field(default_factory=time.time)
    warmup_start_time: Optional[float] = None
    warmup_complete_time: Optional[float] = None
    last_execution_start: Optional[float] = None

    def is_alive(self) -> bool:
        return self.process.is_alive()

    def age_seconds(self) -> float:
        return time.time() - self.spawn_time

    def is_ready(self) -> bool:
        return self.status == ExecutorStatus.IDLE and self.is_alive()

    def mark_busy(self, execution_id: str) -> None:
        self.status = ExecutorStatus.BUSY
        self.current_execution_id = execution_id
        self.last_execution_start = time.time()

    def mark_idle(self) -> None:
        self.status = ExecutorStatus.IDLE
        self.current_execution_id = None
        self.executions_completed += 1

    def mark_dead(self) -> None:
        self.status = ExecutorStatus.DEAD
        self.current_execution_id = None

    def kill(self) -> None:
        if self.is_alive():
            try:
                self.process.terminate()
                self.process.join(timeout=2.0)
                if self.process.is_alive():
                    self.process.kill()
                    self.process.join(timeout=1.0)
            except Exception as e:
                AbstraLogger.error(f"[ExecutorPool] Error killing executor: {e}")
        self.mark_dead()


class ExecutorPool:
    def __init__(
        self,
        config: ExecutorConfig,
        mp_context: MPContext,
        root_path: str,
        server_port: int,
        worker_id: str,
        parent_executions_queue: Optional[Queue] = None,
        verbose: bool = True,
    ):
        self.config = config
        self.mp_context = mp_context
        self.root_path = root_path
        self.server_port = server_port
        self.worker_id = worker_id
        self.parent_executions_queue = parent_executions_queue
        self.verbose = verbose

        self.executors: Dict[str, ExecutorHandle] = {}
        self.execution_to_executor: Dict[str, str] = {}
        self._pending_spawns: List[tuple] = []

        self.lock = RLock()

        self.shutdown_requested = False

        self.metrics = MetricsCollector()

    def start(self) -> None:
        if self.verbose:
            AbstraLogger.warning(
                f"[ExecutorPool] Starting {self.config.total_executors} executors "
                f"with warmup_parallelism={self.config.warmup_parallelism}"
            )

        self._pending_spawns: List[tuple] = [
            (i, i < self.config.min_form_executors)
            for i in range(self.config.total_executors)
        ]

        self.monitor_thread = Thread(target=self._monitor_loop, daemon=True)
        self.monitor_thread.start()

    def shutdown(self) -> None:
        if self.verbose:
            AbstraLogger.warning("[ExecutorPool] Shutting down pool")
        self.shutdown_requested = True

        with self.lock:
            for executor in list(
                self.executors.values()
            ):  # Create snapshot to avoid dictionary modification during iteration
                try:
                    if executor.is_alive():
                        executor.work_queue.put(ShutdownRequest(), timeout=2.0)
                except Exception as e:
                    AbstraLogger.error(
                        f"[ExecutorPool] Error sending shutdown to executor: {e}"
                    )

        shutdown_start = time.time()
        for executor in list(
            self.executors.values()
        ):  # Create snapshot to avoid dictionary modification during iteration
            remaining_time = max(0, 5.0 - (time.time() - shutdown_start))
            executor.process.join(timeout=remaining_time)
            if executor.is_alive():
                executor.kill()

        if self.verbose:
            AbstraLogger.warning("[ExecutorPool] Pool shutdown complete")

    def execute(
        self,
        stage: StageWithFile,
        worker_id: str,
        execution_id: str,
        request,
        connection,
        rabbitmq_params: Optional[RabbitMQParams],
    ) -> ExecutorResponse:
        max_wait = self.config.acquire_timeout_seconds
        start_time = time.time()
        retry_count = 0

        while True:
            executor = self._acquire_executor(stage.type_name, execution_id)

            if executor is not None:
                break

            retry_count += 1

            elapsed = time.time() - start_time
            if elapsed >= max_wait:
                with self.lock:
                    status_summary = {}
                    for e in self.executors.values():
                        status = e.status.value
                        status_summary[status] = status_summary.get(status, 0) + 1

                AbstraLogger.error(
                    f"[ExecutorPool] No executor available after {elapsed:.1f}s wait "
                    f"(execution_id={execution_id}, stage_type={stage.type_name}, "
                    f"executor_status={status_summary})"
                )
                raise RuntimeError(
                    f"No available executor for stage type {stage.type_name} after {elapsed:.1f}s wait. "
                    "All executors are busy or dead."
                )

            time.sleep(0.25)

        try:
            self.metrics.record_execution_started(
                execution_id=execution_id,
                stage_type=stage.type_name,
                executor_id=executor.executor_id,
                wait_time_ms=0.0,
            )

            if not executor.is_alive():
                AbstraLogger.error(
                    f"[ExecutorPool] Executor died before work submission "
                    f"(execution_id={execution_id}, executor_id={executor.executor_id}, pid={executor.pid})"
                )
                executor.mark_dead()
                self.metrics.record_executor_died(executor.executor_id)
                raise RuntimeError(
                    f"Executor {executor.executor_id} died before execution"
                )

            execute_request = ExecuteRequest(
                command=ExecutorCommand.EXECUTE,
                worker_id=worker_id,
                execution_id=execution_id,
                stage=stage,
                request=request,
                connection=connection,
                rabbitmq_params=rabbitmq_params,
            )

            try:
                executor.work_queue.put(execute_request, timeout=5.0)
            except Exception as e:
                AbstraLogger.error(
                    f"[ExecutorPool] Failed to send work request to executor "
                    f"(execution_id={execution_id}, executor_id={executor.executor_id}, error={e})"
                )

                executor.mark_dead()
                self.metrics.record_executor_died(executor.executor_id)
                raise RuntimeError(
                    f"Failed to send work to executor {executor.executor_id}: {e}"
                )

            timeout = self.config.execution_timeout_seconds

            response = None
            wait_start = time.time()

            while time.time() - wait_start < timeout:
                if not executor.is_alive():
                    AbstraLogger.error(
                        f"[ExecutorPool] Executor died while waiting for response "
                        f"(execution_id={execution_id}, executor_id={executor.executor_id}, pid={executor.pid})"
                    )
                    executor.mark_dead()
                    self.metrics.record_executor_died(executor.executor_id)
                    raise RuntimeError(
                        f"Executor {executor.executor_id} died unexpectedly during execution"
                    )

                try:
                    response = executor.response_queue.get(timeout=1.0)
                    break
                except Exception:  # Empty queue
                    continue

            if response is None:
                raise TimeoutError(f"Execution timed out after {timeout}s")

            if response.execution_id != execution_id:
                raise RuntimeError(
                    f"Received response for wrong execution: expected={execution_id}, got={response.execution_id}"
                )

            execution_time_ms = (time.time() - start_time) * 1000
            self.metrics.record_execution_completed(
                execution_id=execution_id,
                success=response.success,
                execution_time_ms=execution_time_ms,
            )

            return response

        except Exception as e:
            AbstraLogger.error(
                f"[ExecutorPool] Error during execution "
                f"(execution_id={execution_id}): {e}"
            )
            raise

        finally:
            self._release_executor(executor, execution_id)

    def kill_execution(self, execution_id: str) -> bool:
        with self.lock:
            executor_id = self.execution_to_executor.get(execution_id)
            if executor_id is None:
                AbstraLogger.warning(
                    f"[ExecutorPool] Cannot kill execution {execution_id}: "
                    "not found in active executions"
                )
                return False

            executor = self.executors.get(executor_id)
            if executor is None or not executor.is_alive():
                AbstraLogger.warning(
                    f"[ExecutorPool] Cannot kill execution {execution_id}: "
                    "executor is dead"
                )
                self.execution_to_executor.pop(execution_id, None)
                return False

            self.execution_to_executor.pop(execution_id, None)
            self.executors.pop(executor_id, None)

        executor.kill()

        is_form_reserved = executor.is_form_reserved
        replacement = self._spawn_executor(is_form_reserved=is_form_reserved)

        with self.lock:
            self.executors[replacement.executor_id] = replacement

        self._start_warmup(replacement)

        return True

    def get_stats(self) -> Dict[str, Any]:
        with self.lock:
            executors = list(self.executors.values())
            active_executions_count = len(self.execution_to_executor)
            metrics_summary = self.metrics.get_summary()

        total = len(executors)
        idle = sum(1 for e in executors if e.status == ExecutorStatus.IDLE)
        busy = sum(1 for e in executors if e.status == ExecutorStatus.BUSY)
        warming_up = sum(1 for e in executors if e.status == ExecutorStatus.WARMING_UP)
        dead = sum(1 for e in executors if e.status == ExecutorStatus.DEAD)
        form_capable_idle = sum(
            1
            for e in executors
            if e.status == ExecutorStatus.IDLE and e.is_form_reserved
        )

        warming_times = []
        slow_warmups = 0
        current_time = time.time()

        for e in executors:
            if e.status == ExecutorStatus.WARMING_UP and e.warmup_start_time:
                warmup_elapsed = current_time - e.warmup_start_time
                warming_times.append(warmup_elapsed)
                if warmup_elapsed > 5.0:  # Consider > 5s as slow
                    slow_warmups += 1

        return {
            **metrics_summary,
            "total_executors": total,
            "idle": idle,
            "busy": busy,
            "warming_up": warming_up,
            "dead": dead,
            "form_capable_idle": form_capable_idle,
            "active_executions": active_executions_count,
            "slow_warmups": slow_warmups,
            "avg_warmup_time": sum(warming_times) / len(warming_times)
            if warming_times
            else 0,
            "max_warmup_time": max(warming_times) if warming_times else 0,
        }

    def can_start_loop(self) -> bool:
        with self.lock:
            executors = list(self.executors.values())

        idle_count = sum(1 for e in executors if e.status == ExecutorStatus.IDLE)

        return idle_count > 0

    def _spawn_executor(self, is_form_reserved: bool) -> ExecutorHandle:
        executor_id = str(uuid4())[:8]

        work_queue = safe_multiprocessing_queue(self.mp_context)
        response_queue = safe_multiprocessing_queue(self.mp_context)

        p = self.mp_context.Process(
            target=executor_main,
            name=f"Executor-{executor_id}",
            args=(
                work_queue,
                response_queue,
                self.root_path,
                self.server_port,
                self.parent_executions_queue,
                self.verbose,
            ),
        )

        p.start()

        executor = ExecutorHandle(
            executor_id=executor_id,
            process=p,
            work_queue=work_queue,
            response_queue=response_queue,
            status=ExecutorStatus.SPAWNING,
            pid=p.pid,
            is_form_reserved=is_form_reserved,
        )

        self.metrics.record_executor_spawned(
            executor_id=executor_id,
            spawn_time=time.time(),
            is_form_reserved=is_form_reserved,
        )

        return executor

    def _start_warmup(self, executor: ExecutorHandle) -> None:
        executor.status = ExecutorStatus.WARMING_UP
        executor.warmup_start_time = time.time()

        if not executor.is_alive():
            AbstraLogger.error(
                f"[ExecutorPool] Process died before warmup (executor_id={executor.executor_id})"
            )
            executor.mark_dead()
            self.metrics.record_executor_died(executor.executor_id)
            return

        try:
            executor.work_queue.put(WarmupRequest(), timeout=5.0)
        except Exception as e:
            AbstraLogger.error(
                f"[ExecutorPool] Failed to start warmup (executor_id={executor.executor_id}): {e}"
            )
            executor.mark_dead()
            self.metrics.record_executor_died(executor.executor_id)

    def _acquire_executor(
        self, stage_type: StageType, execution_id: str
    ) -> Optional[ExecutorHandle]:
        with self.lock:
            idle: List[ExecutorHandle] = []
            for e in self.executors.values():
                if e.status != ExecutorStatus.IDLE:
                    continue
                if not e.is_alive():
                    e.mark_dead()
                    self.metrics.record_executor_died(e.executor_id)
                    continue
                idle.append(e)

            if not idle:
                return None

            form_executors = [e for e in idle if e.is_form_reserved]
            non_form_executors = [e for e in idle if not e.is_form_reserved]

            if stage_type == "form":
                pool = form_executors + non_form_executors
            else:
                pool = non_form_executors

            if not pool:
                return None

            executor = pool[0]

            executor.mark_busy(execution_id)
            self.execution_to_executor[execution_id] = executor.executor_id

            return executor

    def _release_executor(self, executor: ExecutorHandle, execution_id: str) -> None:
        should_recycle = False

        with self.lock:
            self.execution_to_executor.pop(execution_id, None)

            if not executor.is_alive():
                AbstraLogger.warning(
                    f"[ExecutorPool] Executor died during execution "
                    f"(executor_id={executor.executor_id}, pid={executor.pid})"
                )
                executor.mark_dead()
                self.metrics.record_executor_died(executor.executor_id)
                return

            if (
                executor.executions_completed + 1
                >= self.config.max_executions_per_executor
            ):
                should_recycle = True
                self.executors.pop(executor.executor_id, None)
            else:
                executor.mark_idle()

        if should_recycle:
            executor.kill()
            self.metrics.record_executor_died(executor.executor_id)

            is_form_reserved = executor.is_form_reserved
            replacement = self._spawn_executor(is_form_reserved=is_form_reserved)

            with self.lock:
                self.executors[replacement.executor_id] = replacement

            self._start_warmup(replacement)

    def _monitor_loop(self) -> None:
        while not self.shutdown_requested:
            time.sleep(0.1)

            try:
                should_spawn = False
                spawn_info: Optional[tuple] = None

                with self.lock:
                    if self._pending_spawns:
                        warming_count = sum(
                            1
                            for e in self.executors.values()
                            if e.status == ExecutorStatus.WARMING_UP
                        )
                        if warming_count < self.config.warmup_parallelism:
                            spawn_info = self._pending_spawns.pop(0)
                            should_spawn = True

                if should_spawn and spawn_info is not None:
                    idx, is_form_reserved = spawn_info
                    executor = self._spawn_executor(is_form_reserved=is_form_reserved)

                    with self.lock:
                        self.executors[executor.executor_id] = executor

                    if self.verbose:
                        AbstraLogger.warning(
                            f"[ExecutorPool] Spawned executor {idx + 1}/{self.config.total_executors}"
                        )

                    self._start_warmup(executor)

                warming_executors: List[ExecutorHandle] = []
                timed_out_executors: List[ExecutorHandle] = []
                dead_executors: List[ExecutorHandle] = []

                with self.lock:
                    for executor in list(self.executors.values()):
                        if executor.status == ExecutorStatus.WARMING_UP:
                            if executor.current_execution_id is None:
                                warming_executors.append(executor)
                        elif executor.status == ExecutorStatus.DEAD:
                            dead_executors.append(executor)

                for executor in warming_executors:
                    try:
                        response = executor.response_queue.get_nowait()

                        if response.execution_id is not None:
                            executor.response_queue.put(response)
                            continue

                        with self.lock:
                            if response.success:
                                executor.status = ExecutorStatus.IDLE
                                executor.warmup_complete_time = time.time()
                                warmup_duration = executor.warmup_complete_time - (
                                    executor.warmup_start_time or 0
                                )

                                if self.verbose:
                                    AbstraLogger.warning(
                                        f"[ExecutorPool] Warmup completed in {warmup_duration:.3f}s"
                                    )

                                self.metrics.record_executor_warmup(
                                    executor_id=executor.executor_id,
                                    warmup_duration_ms=warmup_duration * 1000,
                                )
                            else:
                                AbstraLogger.error(
                                    f"[ExecutorPool] Warmup failed (executor_id={executor.executor_id})"
                                )
                                executor.mark_dead()
                                self.metrics.record_executor_died(executor.executor_id)
                    except Exception:
                        if executor.warmup_start_time:
                            warmup_elapsed = time.time() - executor.warmup_start_time

                            if warmup_elapsed > self.config.warmup_timeout_seconds:
                                AbstraLogger.error(
                                    f"[ExecutorPool] Warmup timeout (executor_id={executor.executor_id}, "
                                    f"elapsed={warmup_elapsed:.1f}s)"
                                )
                                timed_out_executors.append(executor)

                for executor in timed_out_executors:
                    executor.kill()

                for dead_executor in dead_executors:
                    with self.lock:
                        self.executors.pop(dead_executor.executor_id, None)

                    replacement = self._spawn_executor(
                        is_form_reserved=dead_executor.is_form_reserved
                    )

                    with self.lock:
                        self.executors[replacement.executor_id] = replacement

                    self._start_warmup(replacement)

            except Exception as e:
                AbstraLogger.error(f"[ExecutorPool] Error in monitor loop: {e}")
                AbstraLogger.capture_exception(e)



from abstra.tasks import get_trigger_task, send_task

# You can get the task that triggered this script execution
task = get_trigger_task()

task.complete()

# And also send tasks to the next stages of your workflow
send_task("confirmation", {"message": "Your request was approved"})



import re
import unicodedata

from abstra_internals.settings import Settings

from .base_migration import Migration


class Migration015(Migration):
    @staticmethod
    def target_version() -> str:
        return "15.0"

    @staticmethod
    def source_version() -> str:
        return "14.0"

    def get_all_stages(self):
        forms = [{"type": "form", **x} for x in self.data.get("forms", [])]
        scripts = [{"type": "script", **x} for x in self.data.get("scripts", [])]
        jobs = [{"type": "job", **x} for x in self.data.get("jobs", [])]
        hooks = [{"type": "hook", **x} for x in self.data.get("hooks", [])]
        return forms + scripts + jobs + hooks

    def to_snake_case(self, text: str) -> str:
        lower = text.lower()
        unaccented = (
            unicodedata.normalize("NFD", lower)
            .encode("ascii", "ignore")
            .decode("ascii")
        )
        underscored = re.sub(r"[^a-zA-Z0-9]", "_", unaccented)
        aggregated = re.sub(r"_+", "_", underscored)
        return aggregated.lstrip("_").rstrip("_")

    def _migrate(self) -> None:
        file_to_stages = {}
        for stage in self.get_all_stages():
            file = stage.get("file")
            if file:
                file_to_stages.setdefault(file, []).append(stage)

        new_file_map = {}
        for shared_file, stages in file_to_stages.items():
            if len(stages) <= 1:
                continue

            type_title_to_stages = {}
            for stage in stages:
                key = (stage["type"], self.to_snake_case(stage["title"]))
                type_title_to_stages.setdefault(key, []).append(stage)

            for (stage_type, snake_title), group in type_title_to_stages.items():
                for idx, stage in enumerate(group):
                    serial_suffix = f"_{idx + 1}" if idx > 0 else ""
                    new_filename = f"{stage_type}_{snake_title}{serial_suffix}.py"
                    new_filepath = Settings.root_path / new_filename

                    import_statement = f"import {shared_file[:-3]}"
                    new_filepath.write_text(import_statement, encoding="utf-8")
                    new_file_map[(stage["type"], stage["id"])] = new_filename

        for stage_type, data_key in [
            ("form", "forms"),
            ("script", "scripts"),
            ("job", "jobs"),
            ("hook", "hooks"),
        ]:
            for stage_data in self.data.get(data_key, []):
                key = (stage_type, stage_data["id"])
                if key in new_file_map:
                    stage_data["file"] = new_file_map[key]



import threading
import traceback


class ExceptionAwareThread(threading.Thread):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.creator_trace = traceback.format_stack()
        self.exception = None

    def run(self):
        try:
            super().run()
        except Exception as e:
            self.exception = e

    def join(self, *a, **kw):
        super().join(*a, **kw)
        if self.exception:
            creator_info = "".join(self.creator_trace)
            exception_type = type(self.exception).__name__
            exception_msg = str(self.exception)

            new_exception = RuntimeError(
                f"Exception in thread (created here):\n"
                f"Original exception: {exception_type}: {exception_msg}\n"
                f"Thread creation stack:\n{creator_info}"
            )
            raise new_exception from self.exception



from .base_migration import Migration


class Migration011(Migration):
    @staticmethod
    def target_version() -> str:
        return "11.0"

    @staticmethod
    def source_version() -> str:
        return "10.0"

    def _migrate(self) -> None:
        self.data["language"] = "en"



from typing import List

from abstra_internals.repositories.linter.models import LinterIssue, LinterRule
from abstra_internals.utils.code_check import CodeCheckResult, code_check_all


class TypeCheckingIssue(LinterIssue):
    def __init__(self, result: CodeCheckResult) -> None:
        super().__init__()
        self.result = result
        self.fixes = []


class TypeCheckingRule(LinterRule):
    label = "Type checking"
    type = "bug"

    def find_issues(self) -> List[TypeCheckingIssue]:
        result = code_check_all()

        if result.success:
            return []
        else:
            issue = TypeCheckingIssue(result)
            issue.label = result.stderr if result.stderr else result.stdout
            issue.fixes = []
            return [issue]



import sys
from pathlib import Path


def _is_headless():
    return not sys.stdin.isatty()


def _there_is_abstra_json():
    return (Path.cwd() / "abstra.json").exists()


def _askdirectory():
    ans = input(
        f"The project will be created in the current directory ({Path.cwd()}). Continue? (y/n): "
    )

    if ans.lower() != "y":
        print("\nTry using: 'abstra editor <PATH>' to specify a directory.")
        exit(0)

    return "."


def select_dir() -> str:
    if _there_is_abstra_json() or _is_headless():
        return "."

    asked_dir = _askdirectory()
    if asked_dir != "":
        return asked_dir

    print("No directory selected. Please, select a directory.")
    sys.exit(0)



from typing import Any, List

from abstra_internals.interface.sdk.forms.deprecated.widgets.widget_base import Input


class ListInput(Input):
    type = "list-input"
    schemas: List[Any]
    item_schema: Any
    empty_value = []

    def __init__(self, key: str, item_schema: Any, **kwargs):
        initial_error = kwargs.get("errors", [])
        if not isinstance(initial_error, list):
            initial_error = [initial_error]
        super().__init__(key)
        self.item_schema = item_schema
        self.schemas = []
        self.value = self.empty_value
        self.set_props(dict(item_schema=item_schema, **kwargs))

    def serialize_value(self) -> List:
        return [schema.serialize_value() for schema in self.schemas]

    def set_props(self, props):
        self.min = props.get("min", 0)
        self.max = props.get("max", None)
        self.hint = props.get("hint", None)
        self.add_button_text = props.get("add_button_text", "+")
        self.full_width = props.get("full_width", False)
        self.required = props.get("required", True)
        self.disabled = props.get("disabled", False)
        self.schemas = [self.item_schema.copy() for _ in range(self.min)]
        if props.get("initial_value", None):
            self.set_value(props.get("initial_value", None), set_errors=False)

    def _validate_list_min_max(self) -> List[str]:
        if self.min is not None and len(self.value) < self.min:
            return ["i18n_error_min_list"]
        if self.max is not None and len(self.value) > self.max:
            return ["i18n_error_max_list"]
        return []

    def _get_inputs(self):
        inputs = []
        for schema in self.schemas:
            inputs.extend(schema.get_input_widgets())
        return inputs

    def _validate_each(self) -> List[str]:
        inputs = self._get_inputs()
        for input in inputs:
            input.set_errors()
        if self._has_errors():
            return ["i18n_error_invalid_list_item"]
        return []

    def _has_errors(self) -> bool:
        inputs = self._get_inputs()
        for input in inputs:
            has_errors = len(input.validate()) > 0
            if has_errors:
                return True
        return False

    @property
    def validators(self):
        return (
            super().validators + [self._validate_list_min_max] + [self._validate_each]
        )

    def render(self, ctx: dict):
        schemas = [schema.render(ctx) for schema in self.schemas]
        return {
            "type": self.type,
            "key": self.key,
            "hint": self.hint,
            "errors": self.errors,
            "min": self.min,
            "max": self.max,
            "addButtonText": self.add_button_text,
            "fullWidth": self.full_width,
            "required": self.required,
            "disabled": self.disabled,
            "schemas": schemas,
            "value": self.serialize_value(),
        }

    def set_value(self, value: List, set_errors=False):
        if not self._has_errors():
            self.errors = []
        if not value:
            value = []
        self.value = value
        for idx, value_item in enumerate(value):
            if idx >= len(self.schemas):
                schema = self.item_schema.copy()
                self.schemas.append(schema)
            else:
                schema = self.schemas[idx]
            schema.set_values(value_item)
        self.schemas = self.schemas[: len(value)]

    def parse_value(self, value) -> List:
        parsed_values = []
        for index, value in enumerate(value or self.empty_value):
            if index >= len(self.schemas):
                schema = self.item_schema.copy()
                value = schema.parse_value(value or {})
            else:
                schema = self.schemas[index]
                value = schema.parse_value(value or {})
            parsed_values.append(value)
        return parsed_values



import shutil
import zipfile
from io import BytesIO
from pathlib import Path
from typing import Optional

import requests


def get_default_branch(repo_url: str) -> str:
    """
    Gets the default branch name for a GitHub repository.

    Args:
        repo_url: The GitHub repository URL (e.g., "https://github.com/user/repo")

    Returns:
        The default branch name (e.g., "main", "master")

    Raises:
        requests.RequestException: If the API request fails
        KeyError: If the response doesn't contain the expected data
    """
    parts = repo_url.rstrip("/").split("github.com/")[-1]

    api_url = f"https://api.github.com/repos/{parts}"
    response = requests.get(api_url)
    response.raise_for_status()

    data = response.json()
    return data["default_branch"]


def download_github_repo(
    repo_url: str,
    extract_path: Path,
    branch_name: str = "master",
    target_folder_name: Optional[str] = None,
) -> None:
    """
    Downloads a GitHub repository as a zip file and extracts it to the specified path.

    Args:
        repo_url: The GitHub repository URL (e.g., "https://github.com/user/repo")
        extract_path: The path where the repository will be extracted
        branch_name: The branch name to download (default: "master")
        target_folder_name: Optional target folder name. If provided, moves contents from
                          "repo-branch" to this folder. If None, keeps the extracted folder as-is.

    Raises:
        requests.RequestException: If the download fails
        zipfile.BadZipFile: If the downloaded file is not a valid zip
        Exception: For other errors during extraction or file operations
    """
    response = requests.get(f"{repo_url}/archive/refs/heads/{branch_name}.zip")
    response.raise_for_status()

    with zipfile.ZipFile(BytesIO(response.content)) as z:
        z.extractall(extract_path)

    if target_folder_name:
        repo_name = repo_url.rstrip("/").split("/")[-1]
        extracted_folder = extract_path / f"{repo_name}-{branch_name}"
        target_folder = extract_path / target_folder_name

        if extracted_folder.exists():
            target_folder.mkdir(exist_ok=True)
            for item in extracted_folder.iterdir():
                dest = target_folder / item.name
                if dest.exists():
                    if dest.is_dir():
                        shutil.rmtree(dest)
                    else:
                        dest.unlink()
                shutil.move(str(item), str(dest))
            shutil.rmtree(extracted_folder)



from abstra.tasks import get_tasks, send_task

# Abstra Jobs are a simple way to schedule Python recurring tasks
# Setup the Schedule at the "Settings" tab

print("🕒 Job is running...")

# You can get all the pending tasks, waiting to be processes in this stage
tasks = get_tasks()

processed_tasks = 0
for task in tasks:
    # Do any processing here
    processed_tasks += 1
    task.complete()

send_task("finished", {"proccessed_tasks": processed_tasks})



import flask

from abstra_internals.controllers.main import MainController
from abstra_internals.repositories.project.project import JobStage
from abstra_internals.usage import editor_usage
from abstra_internals.utils import is_it_true


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_jobs", __name__)

    @bp.get("/<path:id>")
    @editor_usage
    def _get_job(id: str):
        job = controller.get_job(id)
        if not job:
            flask.abort(404)
        return job.editor_dto

    @bp.get("/")
    @editor_usage
    def _get_jobs():
        return [f.editor_dto for f in controller.get_jobs()]

    @bp.post("/")
    @editor_usage
    def _create_job():
        data = flask.request.json
        if not data:
            flask.abort(400)
        title = data.get("title")
        file = data.get("file")
        if not title or not file:
            flask.abort(400)
        workflow_position = data.get("position", (0, 0))
        id = data.get("id", None)
        job = controller.create_job(title, file, workflow_position, id)
        return job.editor_dto

    @bp.put("/<path:id>")
    @editor_usage
    def _update_stage(id: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        job = controller.update_stage(id, data)
        if isinstance(job, JobStage):
            return job.editor_dto
        else:
            return None

    @bp.delete("/<path:id>")
    @editor_usage
    def _delete_job(id: str):
        remove_file = flask.request.args.get(
            "remove_file", default=False, type=is_it_true
        )
        controller.delete_stage(id, remove_file)
        return {"success": True}

    @bp.post("/<path:id>/run")
    @editor_usage
    def _run_job(id: str):
        status = controller.get_job_status(id)
        if status == "not_found":
            flask.abort(404)

        if status == "disabled":
            return {"status": "disabled"}

        return controller.run_job(id)

    return bp



from typing import TYPE_CHECKING, Callable, Optional

from abstra_internals.interface.sdk.forms.deprecated.widgets.widget_base import Input

if TYPE_CHECKING:
    from abstra_internals.interface.sdk.forms.deprecated.page import Page


class Reactive(Input):
    type = "reactive"
    page: Optional["Page"]
    callback: Callable[[dict], Optional["Page"]]

    def __init__(self, callback: Callable, **kwargs):
        self.set_props(dict(callback=callback, **kwargs))
        self.page = None

    def has_errors(self):
        return self.page.has_errors() if self.page else False

    def _are_pages_different(self, page1: Optional["Page"], page2: Optional["Page"]):
        if page1 is None:
            return page2 is not None

        if page2 is None:
            return True

        return not page1.is_equal_to(page2)

    def render(self, ctx: dict):
        payload = {}
        payload.update(self.value)
        payload.update(ctx)
        new_page = self.callback(payload)

        if self._are_pages_different(self.page, new_page):
            self.page = new_page

            if self.page is not None:
                self.page.set_values(self.value)

        if self.page is None:
            return []

        return self.page.render(ctx)

    def set_value(self, value, set_errors=False):
        self.value = value
        if not hasattr(self, "page") or self.page is None:
            try:
                self.page = self.callback({})
            except Exception:
                self.page = None

        if self.page:
            self.page.set_values(value)

    def serialize_value(self):
        return self.page.serialize_value() if self.page else {}

    def set_props(self, props):
        self.callback = props.get("callback")
        self.set_value(props.get("initial_value", {}))
        self.page = None

    def set_errors(self):
        if self.page:
            self.page.set_errors()

    def parse_value(self, value):
        if self.page:
            parsed = self.page.parse_value(value)
            return parsed
        else:
            return {}



from unittest import TestCase
from unittest.mock import ANY

from tests.fixtures import clear_dir, init_dir

from .format import tables_snapshot_from_csv, tables_snapshot_to_csv


class TestFromCsv(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_from_csv_rn(self):
        csv_path = self.root / "test.csv"
        lines = (
            "name,age,city\r\n"
            "Alice,30,New York\r\n"
            "Bob,25,Los Angeles\r\n"
            "Charlie,35,Chicago\r\n"
            "David,40,Houston\r\n"
        )
        csv_path.write_text(lines, encoding="utf-8")
        restore_request = tables_snapshot_from_csv(csv_path)
        self.assertEqual(
            restore_request.to_dict(),
            dict(
                tables=[
                    dict(
                        id=ANY,
                        name="test",
                        data=[
                            dict(name="Alice", age="30", city="New York"),
                            dict(name="Bob", age="25", city="Los Angeles"),
                            dict(name="Charlie", age="35", city="Chicago"),
                            dict(name="David", age="40", city="Houston"),
                        ],
                        columns=[
                            dict(
                                id=ANY,
                                primaryKey=False,
                                name="name",
                                type="text",
                                unique=False,
                                nullable=False,
                            ),
                            dict(
                                id=ANY,
                                primaryKey=False,
                                name="age",
                                type="text",
                                unique=False,
                                nullable=False,
                            ),
                            dict(
                                id=ANY,
                                primaryKey=False,
                                name="city",
                                type="text",
                                unique=False,
                                nullable=False,
                            ),
                        ],
                    )
                ]
            ),
        )

    def test_from_csv_n(self):
        csv_path = self.root / "test.csv"
        lines = (
            "name,age,city\n"
            "Alice,30,New York\n"
            "Bob,25,Los Angeles\n"
            "Charlie,35,Chicago\n"
            "David,40,Houston\n"
        )
        csv_path.write_text(lines, encoding="utf-8")
        restore_request = tables_snapshot_from_csv(csv_path)
        self.assertEqual(
            restore_request.to_dict(),
            dict(
                tables=[
                    dict(
                        id=ANY,
                        name="test",
                        data=[
                            dict(name="Alice", age="30", city="New York"),
                            dict(name="Bob", age="25", city="Los Angeles"),
                            dict(name="Charlie", age="35", city="Chicago"),
                            dict(name="David", age="40", city="Houston"),
                        ],
                        columns=[
                            dict(
                                id=ANY,
                                primaryKey=False,
                                name="name",
                                type="text",
                                unique=False,
                                nullable=False,
                            ),
                            dict(
                                id=ANY,
                                primaryKey=False,
                                name="age",
                                type="text",
                                unique=False,
                                nullable=False,
                            ),
                            dict(
                                id=ANY,
                                primaryKey=False,
                                name="city",
                                type="text",
                                unique=False,
                                nullable=False,
                            ),
                        ],
                    )
                ]
            ),
        )

    def test_from_csv_with_empty_lines(self):
        csv_path = self.root / "test.csv"
        lines = "name,age,city"
        csv_path.write_text(lines, encoding="utf-8")
        restore_request = tables_snapshot_from_csv(csv_path)
        self.assertEqual(
            restore_request.to_dict(),
            dict(
                tables=[
                    dict(
                        id=ANY,
                        name="test",
                        data=[],
                        columns=[
                            dict(
                                id=ANY,
                                primaryKey=False,
                                name="name",
                                type="text",
                                unique=False,
                                nullable=False,
                            ),
                            dict(
                                id=ANY,
                                primaryKey=False,
                                name="age",
                                type="text",
                                unique=False,
                                nullable=False,
                            ),
                            dict(
                                id=ANY,
                                primaryKey=False,
                                name="city",
                                type="text",
                                unique=False,
                                nullable=False,
                            ),
                        ],
                    )
                ]
            ),
        )

    def test_from_csv_detect_delimiter(self):
        self.maxDiff = None
        csv_path = self.root / "test.csv"
        lines = (
            "name;age;city\n"
            "Alice;30;New York\n"
            "Bob;25;Los Angeles\n"
            "Charlie;35;Chicago\n"
            "David;40;Houston\n"
        )
        csv_path.write_text(lines, encoding="utf-8")
        restore_request = tables_snapshot_from_csv(csv_path)
        self.assertEqual(
            restore_request.to_dict(),
            dict(
                tables=[
                    dict(
                        id=ANY,
                        name="test",
                        data=[
                            dict(name="Alice", age="30", city="New York"),
                            dict(name="Bob", age="25", city="Los Angeles"),
                            dict(name="Charlie", age="35", city="Chicago"),
                            dict(name="David", age="40", city="Houston"),
                        ],
                        columns=[
                            dict(
                                id=ANY,
                                primaryKey=False,
                                name="name",
                                type="text",
                                unique=False,
                                nullable=False,
                            ),
                            dict(
                                id=ANY,
                                primaryKey=False,
                                name="age",
                                type="text",
                                unique=False,
                                nullable=False,
                            ),
                            dict(
                                id=ANY,
                                primaryKey=False,
                                name="city",
                                type="text",
                                unique=False,
                                nullable=False,
                            ),
                        ],
                    )
                ]
            ),
        )

    def test_to_csv(self):
        csv_path = self.root / "test.csv"
        lines = (
            "name,age,city\r\n"
            "Alice,30,New York\r\n"
            "Bob,25,Los Angeles\r\n"
            "Charlie,35,Chicago\r\n"
            "David,40,Houston\r\n"
        )
        csv_path.write_text(lines, encoding="utf-8")
        restore_request = tables_snapshot_from_csv(csv_path)
        table_name, csv_content = tables_snapshot_to_csv(restore_request)
        self.assertEqual(table_name, "test")
        self.assertEqual(csv_content, lines)



import os
import ssl
import subprocess
import sys
import threading

import certifi
from dotenv import load_dotenv
from werkzeug.serving import make_server

from abstra_internals.cloud_api import connect_tunnel
from abstra_internals.controllers.codebase_events import CodebaseEventController
from abstra_internals.controllers.execution.consumer import ConsumerController
from abstra_internals.controllers.main import MainController
from abstra_internals.environment import EDITOR_MODE, HOST, RABBITMQ_CONNECTION_URI
from abstra_internals.interface.cli.messages import serve_message
from abstra_internals.logger import AbstraLogger
from abstra_internals.logs_watcher import LogsWatcher, on_logs_update
from abstra_internals.repositories.consumer import EditorConsumer
from abstra_internals.repositories.factory import (
    build_editor_repositories,
    build_web_editor_repositories,
    get_mp_context_repository,
)
from abstra_internals.repositories.producer import (
    LocalProducerRepository,
)
from abstra_internals.server.apps import get_local_app
from abstra_internals.services.file_watcher import FileWatcher
from abstra_internals.settings import Settings
from abstra_internals.stdio_patcher import StdioPatcher
from abstra_internals.tasks_watcher import TasksWatcher, on_tasks_update
from abstra_internals.utils.browser import background_open_editor
from abstra_internals.utils.multiprocessing import safe_multiprocessing_queue
from abstra_internals.version import check_latest_version


def start_consumer(controller: MainController, debug_mode: bool = False):
    if isinstance(controller.producer_repository, LocalProducerRepository):
        consumer = EditorConsumer(controller.producer_repository.queue)
        consumer_controller = ConsumerController(
            controller, consumer, debug_mode=debug_mode
        )

        th = threading.Thread(
            daemon=True,
            name="start_consumer::EditorConsumer",
            target=consumer_controller.start_loop,
        )

        th.start()
        return consumer, th, consumer_controller

    raise ValueError("Invalid producer repository")


def ensure_certificates():
    try:
        cafile = certifi.where()
        if not os.path.isfile(cafile):
            raise FileNotFoundError(f"Certifi CA file not found at {cafile}")
        ssl.create_default_context(cafile=cafile)
    except Exception:
        print(
            "SSL certificate validation failed. Attempting to restore certificates..."
        )
        try:
            subprocess.run(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--force-reinstall",
                    "certifi",
                ],
                check=True,
            )
            import certifi as certifi_reinstalled

            cafile_re = certifi_reinstalled.where()
            if os.path.isfile(cafile_re):
                print(f"Certificates restored at: {cafile_re}")
            else:
                print(
                    f"Certifi CA file still missing at: {cafile_re}. Please check your Python environment."
                )
        except Exception as update_e:
            print(f"Failed to restore certificates: {update_e}")


def editor(headless: bool, verbose: bool = False, debug_mode: bool = False):
    ensure_certificates()

    load_dotenv(Settings.root_path / ".env")

    serve_message()
    check_latest_version()
    AbstraLogger.init("local" if EDITOR_MODE == "local" else "cloud")

    # Determine if we should use RabbitMQ based on EDITOR_MODE and RABBITMQ_CONNECTION_URI
    # Web editor with workers: EDITOR_MODE=web + RABBITMQ_CONNECTION_URI set
    # Web editor without workers: EDITOR_MODE=web + no RABBITMQ_CONNECTION_URI (legacy mode)
    # Local editor: EDITOR_MODE=local
    is_web_editor = EDITOR_MODE == "web"
    use_rabbitmq_workers = is_web_editor and RABBITMQ_CONNECTION_URI is not None

    AbstraLogger.info(
        f"[Editor] Configuration: EDITOR_MODE={EDITOR_MODE}, RABBITMQ_CONNECTION_URI={'SET' if RABBITMQ_CONNECTION_URI else 'NOT SET'}"
    )

    if use_rabbitmq_workers:
        AbstraLogger.info(
            "[Editor] Running in web editor mode with RabbitMQ workers (isolated execution)"
        )
        assert RABBITMQ_CONNECTION_URI is not None
        repositories = build_web_editor_repositories(RABBITMQ_CONNECTION_URI)
    else:
        if is_web_editor:
            AbstraLogger.info(
                "[Editor] Running in web editor mode without workers (legacy mode)"
            )
        else:
            AbstraLogger.info("[Editor] Running in local editor mode")
        mp_context = get_mp_context_repository()
        local_queue = safe_multiprocessing_queue(mp_context.get_context())
        repositories = build_editor_repositories(local_queue)

    main_controller = MainController(repositories)
    main_controller.reset_repositories()
    StdioPatcher.apply(main_controller)

    codebase_event_controller = CodebaseEventController(repositories)
    watcher = FileWatcher(
        [
            codebase_event_controller.reload_env,
            codebase_event_controller.reload_modules,
            codebase_event_controller.lint_files,
            codebase_event_controller.broadcast_changes,
        ]
    )
    watcher.start()

    logs_watcher = LogsWatcher([on_logs_update])
    logs_watcher.start()

    tasks_watcher = TasksWatcher([on_tasks_update])
    tasks_watcher.start()

    if not is_web_editor:
        start_consumer(main_controller, debug_mode=debug_mode)

    app = get_local_app(main_controller)
    server = make_server(host=HOST, port=Settings.server_port, threaded=True, app=app)

    if not headless:
        background_open_editor()

    connect_tunnel(verbose=verbose)

    server.serve_forever()



import json
from abc import ABC, abstractmethod

from abstra_internals.cloud_api import get_session_path, get_tunnel_secret_key
from abstra_internals.cloud_api.http_client import HTTPClient
from abstra_internals.contracts_generated import (
    CloudApiCliAiV2AbortRequest,
    CloudApiCliAiV2ConversationPostRequest,
    CloudApiCliAiV2ConversationPostResponse,
    CloudApiCliAiV2PromptPostRequest,
    CloudApiCliAiV2StreamRequest,
)
from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import REQUEST_TIMEOUT


class AIRepository(ABC):
    def __init__(self, client: "HTTPClient") -> None:
        self.client = client

    @abstractmethod
    def prompt(self, prompt_request_body: CloudApiCliAiV2PromptPostRequest):
        raise NotImplementedError()

    @abstractmethod
    def parse_document(self, model: str, file_content: bytes, mime_type: str):
        raise NotImplementedError()

    @abstractmethod
    def get_ai_messages(
        self,
        req: CloudApiCliAiV2StreamRequest,
    ):
        raise NotImplementedError()

    @abstractmethod
    def get_history(self, headers: dict, limit: int, offset: int):
        raise NotImplementedError()

    @abstractmethod
    def create_thread(self, headers: dict) -> CloudApiCliAiV2ConversationPostResponse:
        raise NotImplementedError()

    @abstractmethod
    def delete_thread(self, headers: dict, thread_id: str):
        raise NotImplementedError()

    @abstractmethod
    def abort_thread(self, headers: dict, thread_id: str):
        raise NotImplementedError()

    @abstractmethod
    def start_conversation(self, secret_key: str, tunnel_session_path: str):
        raise NotImplementedError()


class ProductionAIRepository(AIRepository):
    def prompt(self, prompt_request_body: CloudApiCliAiV2PromptPostRequest):
        response = self.client.post(
            endpoint="/ai-v2/prompt",
            json=prompt_request_body.to_dict(),
        )

        try:
            response = response.json()
            return response
        except json.JSONDecodeError:
            raise Exception(f"Error parsing JSON: {response.text}")

    def parse_document(self, model: str, file_content: bytes, mime_type: str):
        response = self.client.post(
            endpoint=f"/ai-v2/parse-document/{model}",
            headers={"Content-Type": mime_type},
            data=file_content,
        )
        response.raise_for_status()

        return response.json()

    def get_ai_messages(
        self,
        req: CloudApiCliAiV2StreamRequest,
    ):
        raise NotImplementedError()

    def get_history(self, headers: dict, limit: int, offset: int):
        raise NotImplementedError()

    def create_thread(self, headers: dict):
        raise NotImplementedError()

    def abort_thread(self, headers: dict, thread_id: str):
        raise NotImplementedError()

    def start_conversation(self, secret_key: str, tunnel_session_path: str):
        raise NotImplementedError(
            "This method is not implemented in ProductionAIRepository."
        )

    def delete_thread(self, headers: dict, thread_id: str):
        raise NotImplementedError(
            "This method is not implemented in ProductionAIRepository."
        )

    def get_user_info(self):
        raise NotImplementedError(
            "This method is not implemented in ProductionAIRepository."
        )


class LocalAIRepository(AIRepository):
    def prompt(self, prompt_request_body: CloudApiCliAiV2PromptPostRequest):
        response = self.client.post("/ai-v2/prompt", json=prompt_request_body.to_dict())

        response.raise_for_status()

        response = response.json()
        return response

    def parse_document(self, model: str, file_content: bytes, mime_type: str):
        headers = resolve_headers()
        if headers is None:
            raise Exception("You must be logged in to use AI")
        response = self.client.post(
            f"/ai-v2/parse-document/{model}",
            headers={**headers, "Content-Type": mime_type},
            data=file_content,
        )
        response.raise_for_status()

        return response.json()

    def get_ai_messages(
        self,
        req: CloudApiCliAiV2StreamRequest,
    ):
        url = "/ai-v2/stream"
        body = req.to_dict()
        headers = resolve_headers()
        response = self.client.post(url, headers=headers, json=body, stream=True)
        if response.status_code != 200:
            response.raise_for_status()

        return response.iter_content(chunk_size=None)

    def get_history(self, headers: dict, limit: int, offset: int):
        url = "/ai-v2/history"
        r = self.client.get(
            url,
            headers=headers,
            params={"limit": limit, "offset": offset},
            timeout=REQUEST_TIMEOUT,
        )
        r.raise_for_status()
        return r.json()

    def create_thread(self, headers: dict):
        url = "/ai-v2/conversation"
        r = self.client.post(
            url,
            headers=headers,
            timeout=REQUEST_TIMEOUT,
            json=CloudApiCliAiV2ConversationPostRequest(
                tunnel_session_path=get_session_path(),
                secret_key=get_tunnel_secret_key(),
            ).to_dict(),
        )
        r.raise_for_status()
        return CloudApiCliAiV2ConversationPostResponse.from_dict(r.json())

    def abort_thread(self, headers: dict, thread_id: str):
        response = self.client.post(
            endpoint="/ai-v2/abort",
            headers=headers,
            json=CloudApiCliAiV2AbortRequest(conversation_id=thread_id).to_dict(),
        )
        response.raise_for_status()

    def start_conversation(self, secret_key: str, tunnel_session_path: str):
        """
        Start a new conversation with the AI.

        This method initializes a new conversation thread with the AI service.

        Args:
            secret_key (str): The secret key for authentication.
            tunnel_session_path (str): The session path for the tunnel.

        Returns:
            dict: The response containing the conversation details.
        """
        url = "/ai-v2/conversation"
        body = CloudApiCliAiV2ConversationPostRequest(
            tunnel_session_path=tunnel_session_path,
            secret_key=secret_key,
        ).to_dict()
        response = self.client.post(url, json=body)
        response.raise_for_status()
        return response.json()

    def delete_thread(self, headers: dict, thread_id: str):
        """
        Delete a conversation thread.

        Args:
            headers (dict): The headers for the request, including authentication.
            thread_id (str): The ID of the thread to delete.

        Returns:
            None: The method does not return any value.
        """
        print(f"Deleting thread with ID: {thread_id}")
        url = f"/ai-v2/conversation/{thread_id}"
        response = self.client.delete(url, headers=headers)
        response.raise_for_status()



import os
from typing import List

from abstra_internals.repositories.linter.models import LinterRule

from .big_py_files import BigPyFiles
from .conflicting_name import ConflictingName
from .conflicting_path import ConflictingPath
from .deprecated_functions import DeprecatedFunctionUsage
from .duplicate_package_in_requirements import DuplicatePackagesInRequirements
from .env_in_bundle import EnvInBundle
from .file_outside_project import FileOutsideProjectRoot
from .missing_abstra_in_requirements import MissingAbstraInRequirements
from .missing_entrypoint import MissingEntrypoint
from .missing_env import MissingEnv
from .missing_packages_in_requirements import MissingPackagesInRequirements
from .new_version_of_abstra_available import NewVersionOfAbstraAvailable
from .psycopg2 import Psycopg2MustBeBinary
from .syntax_errors import SyntaxErrors

# from .type_checking import TypeCheckingRule
from .venv_in_bundle import VenvInBundle

core_rules: List[LinterRule] = [
    EnvInBundle(),
    VenvInBundle(),
    MissingPackagesInRequirements(),
    SyntaxErrors(),
    MissingEntrypoint(),
    MissingAbstraInRequirements(),
    DuplicatePackagesInRequirements(),
    MissingEnv(),
    ConflictingPath(),
    Psycopg2MustBeBinary(),
    ConflictingName(),
    DeprecatedFunctionUsage(),
    BigPyFiles(),
    FileOutsideProjectRoot(),
    # TypeCheckingRule(),
]

conditional_rules: List[LinterRule] = []

if not os.getenv("ABSTRA_RUNNING_IN_BUNDLED_APP"):
    conditional_rules = [NewVersionOfAbstraAvailable()]

rules: List[LinterRule] = core_rules + conditional_rules



import json
import os
from time import sleep
from typing import TYPE_CHECKING, List, Optional, Union

from pydantic import BaseModel

from abstra.logging import logger

from .drive_commands import commands
from .json_schema import literal, object

if TYPE_CHECKING:
    from selenium.webdriver.remote.webdriver import WebDriver


def command_schema(name: str, params: dict, description: str):
    return {
        "type": "object",
        "description": description,
        "properties": {
            "name": literal(name),
            "params": params,
        },
        "required": ["name", "params"],
    }


class ReActStep(BaseModel):
    thought: Optional[str] = None
    action: Optional[str] = None
    observation: Optional[str] = None


class AiState(BaseModel):
    query: Optional[str] = None
    steps: List[ReActStep] = []
    current_step: int = 0

    @classmethod
    def start(cls, query: str):
        return cls(query=query, steps=[], current_step=1)

    def add_thought(self, thought: str):
        if len(self.steps) == 0 or self.steps[-1].thought is not None:
            self.steps.append(ReActStep(thought=thought))
        else:
            self.steps[-1].thought = thought

    def add_action(self, action: str):
        if len(self.steps) == 0:
            self.steps.append(ReActStep(action=action))
        else:
            self.steps[-1].action = action

    def add_observation(self, observation: str):
        if len(self.steps) > 0:
            self.steps[-1].observation = observation

    def to_prompt(self) -> str:
        trajectory_parts = []
        for i, step in enumerate(self.steps, 1):
            if step.thought:
                trajectory_parts.append(f"Thought {i}: {step.thought}")
            if step.action:
                trajectory_parts.append(f"Action {i}: {step.action}")
            if step.observation:
                trajectory_parts.append(f"Observation {i}: {step.observation}")

        current_trajectory = (
            "\n".join(trajectory_parts) if trajectory_parts else "No actions taken yet."
        )

        # Determine what the AI should do next based on the last step
        last_step = self.steps[-1] if self.steps else None
        next_instruction = ""

        if not self.steps:
            next_instruction = "FIRST STEP: Think about what URL or starting point you need to begin with based on the task, then take an appropriate action"
        elif last_step and last_step.thought and not last_step.action:
            next_instruction = "You just had a thought. Now you MUST take an ACTION. Choose one: get, findElement, findElements, clickElement, or return. Do NOT think again."
        elif last_step and last_step.action and not last_step.observation:
            next_instruction = "Waiting for action result..."
        elif last_step and last_step.observation:
            next_instruction = "Based on the observation, think briefly about what to do next OR take another action to continue."

        # Generate available actions dynamically from commands
        action_descriptions = []
        for cmd_class in commands:
            action_descriptions.append(f"- {cmd_class.__name__}: {cmd_class.__doc__}")

        actions_text = "\n".join(action_descriptions)

        return f"""You are an AI assistant that uses ReAct (Reasoning + Acting) to solve web browsing tasks.

Task: {self.query}

IMPORTANT: You must alternate between thinking and acting. After a thought, you MUST take an action. Do not think multiple times in a row.

Available actions:
{actions_text}
- return: Provide the final answer when you have collected the needed information

Tips for web browsing tasks:
- Use getPageSource to get page HTML and search for specific content
- Use w3cExecuteScript with JavaScript like "return document.body.innerText" to get all visible text
- Navigate to relevant URLs based on your task requirements
- Look for specific sections, elements, or content patterns relevant to your task

Current progress:
{current_trajectory}

{next_instruction}"""


def _normalize_return_params(return_params: Union[dict, BaseModel, None]) -> dict:
    """
    Convert return_params to a dict format for use in JSON schema.

    Args:
        return_params: Either a dict (JSON schema) or BaseModel class/instance

    Returns:
        Dict representing the JSON schema
    """
    if return_params is None:
        return object(
            {
                "done": {
                    "type": "boolean",
                    "description": "Indicates if the task is completed successfully",
                }
            }
        )
    elif isinstance(return_params, dict):
        return return_params
    elif isinstance(return_params, type) and issubclass(return_params, BaseModel):
        # It's a BaseModel class
        return return_params.model_json_schema()
    elif isinstance(return_params, BaseModel):
        # It's a BaseModel instance
        return return_params.model_json_schema()
    else:
        raise ValueError(
            f"return_params must be a dict or BaseModel, got {type(return_params)}"
        )


def experimental_prompt_browser(
    query: str,
    response_format: Union[dict, BaseModel, None] = None,
    *,
    driver: Optional["WebDriver"] = None,
    verbose: bool = False,
) -> dict:
    """
    Main browse function using ReAct framework for AI-powered web browsing.

    Args:
        query (str): The query/task to accomplish through web browsing
        response_format (Optional[Union[dict, pydantic.main.BaseModel]]): JSON schema dict or Pydantic BaseModel defining the expected return format
        driver (Optional[WebDriver]): Optional WebDriver instance to use. If not provided, a new one will be created
        verbose (bool): Enable verbose logging for debugging purposes

    Returns:
        Dict containing the results of the browsing task
    """
    from selenium import webdriver

    from abstra.ai import prompt

    ABSTRA_ENVIRONMENT = os.getenv("ABSTRA_ENVIRONMENT")
    ABSTRA_SELENIUM_URL = os.getenv("ABSTRA_SELENIUM_URL")

    # Normalize return_params to dict format
    response_format = _normalize_return_params(response_format)

    if verbose:
        logger.debug(
            f"[VERBOSE] Starting experimental_prompt_browser with query: {query}"
        )
        logger.debug(f"[VERBOSE] Response format: {response_format}")

    if driver is not None:
        if verbose:
            logger.debug("[VERBOSE] Using provided WebDriver")
    elif ABSTRA_ENVIRONMENT == "production" and ABSTRA_SELENIUM_URL is not None:
        if verbose:
            logger.debug(
                f"[VERBOSE] Creating remote WebDriver with URL: {ABSTRA_SELENIUM_URL}"
            )
        options = webdriver.ChromeOptions()
        options.add_argument("--no-sandbox")
        driver = webdriver.Remote(command_executor=ABSTRA_SELENIUM_URL, options=options)
    else:
        if verbose:
            logger.debug("[VERBOSE] Creating local Chrome WebDriver")
        driver = webdriver.Chrome()

    state = AiState.start(query)
    max_steps = 25  # Increased for more complex tasks
    step_count = 0

    if verbose:
        logger.debug(f"[VERBOSE] Starting ReAct loop with max_steps: {max_steps}")

    try:
        while step_count < max_steps:
            if verbose:
                logger.debug(f"\n[VERBOSE] === Step {step_count + 1}/{max_steps} ===")

            try:
                # Create the format schema for ReAct
                format_schema = {
                    "command": {
                        "oneOf": [
                            # Thought command
                            {
                                "type": "object",
                                "description": "Think about what to do next, analyze the situation, or reason about the task",
                                "properties": {
                                    "name": literal("thought"),
                                    "params": {
                                        "type": "object",
                                        "properties": {
                                            "reasoning": {
                                                "type": "string",
                                                "description": "Your reasoning or thought process",
                                            }
                                        },
                                        "required": ["reasoning"],
                                    },
                                },
                                "required": ["name", "params"],
                            }
                        ]
                        + [
                            # Action commands (existing selenium commands)
                            command_schema(
                                cls.__name__, cls.model_json_schema(), cls.__doc__ or ""
                            )
                            for cls in commands
                        ]
                        + [
                            # Return command
                            command_schema(
                                "return",
                                _normalize_return_params(response_format),
                                "Use this when you have gathered enough information to answer the user's question",
                            )
                        ],
                    }
                }

                if verbose:
                    logger.debug("[VERBOSE] Sending prompt to AI...")
                    logger.debug(
                        f"[VERBOSE] Current state prompt:\n{state.to_prompt()}"
                    )

                res = prompt(state.to_prompt(), format=format_schema)  # type: ignore

                if verbose:
                    logger.debug(f"[VERBOSE] AI response: {res}")

                if not res or "command" not in res:
                    if verbose:
                        logger.debug(
                            "[VERBOSE] Invalid response from AI, skipping step"
                        )
                    step_count += 1
                    continue

                # Handle command parsing more robustly
                try:
                    if isinstance(res["command"], dict):
                        command = res["command"]
                    elif isinstance(res["command"], str):
                        command = json.loads(res["command"])
                    else:
                        if verbose:
                            logger.debug(
                                f"[VERBOSE] Unexpected command type: {type(res['command'])}"
                            )
                        step_count += 1
                        continue
                except json.JSONDecodeError as e:
                    logger.warning("Failed to parse command JSON", exc_info=e)
                    if verbose:
                        logger.debug(f"[VERBOSE] JSON decode error: {e}")
                    step_count += 1
                    continue

                if verbose:
                    logger.debug(f"[VERBOSE] Parsed command: {command}")

                if command["name"] == "thought":
                    # Handle thought step
                    thought_text = command["params"]["reasoning"]

                    if verbose:
                        logger.debug(f"[VERBOSE] Processing thought: {thought_text}")

                    # Check if we have too many consecutive thoughts without action
                    recent_thoughts = 0
                    for step in reversed(state.steps):
                        if step.thought and not step.action:
                            recent_thoughts += 1
                        else:
                            break

                    if recent_thoughts >= 3:
                        # Force an action instead of allowing more thoughts
                        if verbose:
                            logger.debug(
                                f"[VERBOSE] Too many consecutive thoughts ({recent_thoughts}), forcing action"
                            )
                        # Default to thinking about the next action if no clear starting point
                        action_description = "thought with params: {'reasoning': 'I need to determine the appropriate starting point or URL for this task'}"
                        state.add_thought(
                            "I need to determine the appropriate starting point or URL for this task"
                        )
                    else:
                        state.add_thought(thought_text)

                elif command["name"] == "return":
                    # Handle return - end the process
                    if verbose:
                        logger.debug(
                            f"[VERBOSE] AI returned final result: {command['params']}"
                        )
                    return command["params"]

                else:
                    # Handle action step
                    action_name = command["name"]
                    action_params = command["params"]
                    action_description = f"{action_name} with params: {action_params}"

                    if verbose:
                        logger.debug(
                            f"[VERBOSE] Executing action: {action_description}"
                        )

                    state.add_action(action_description)

                    # Execute the selenium command
                    try:
                        result = driver.execute(
                            command["name"], params=command["params"]
                        )
                        observation = f"Successfully executed {action_name}. Result: {str(result)[:500]}..."
                        if verbose:
                            logger.debug(f"[VERBOSE] Action succeeded: {observation}")
                    except Exception as action_error:
                        error_str = str(action_error)
                        observation = (
                            f"Failed to execute {action_name}. Error: {error_str}"
                        )

                        if verbose:
                            logger.debug(f"[VERBOSE] Action failed: {observation}")

                        # If it's an element ID error, suggest alternative approaches
                        if "'id'" in error_str and action_name == "getElementText":
                            observation += " (Hint: Try using getPageSource or w3cExecuteScript to extract text content instead)"

                    state.add_observation(observation)

            except Exception as e:
                error_msg = f"Error in ReAct step: {e}"
                if verbose:
                    logger.debug(f"[VERBOSE] Exception in step: {error_msg}")
                state.add_observation(error_msg)

            step_count += 1
            sleep(0.5)  # Small delay between steps

        # If we reach max steps, try to summarize findings
        if verbose:
            logger.debug(
                f"[VERBOSE] Reached maximum steps ({max_steps}), attempting to summarize findings"
            )

        # Create a summary from all observations
        observations = [step.observation for step in state.steps if step.observation]
        observations_text = "\n".join(
            [f"Step {i + 1}: {obs}" for i, obs in enumerate(observations)]
        )

        if verbose:
            logger.debug(f"[VERBOSE] Total observations collected: {len(observations)}")

        # Ask the AI to summarize findings from all observations
        summary_prompt = f"""
        Based on all the observations from the browsing session, please summarize what was found:
        
        Original question: {query}
        
        All observations:
        {observations_text}
        
        Please extract and summarize the key findings in a structured way, even if the task is incomplete.
        Return in this format: {json.dumps(response_format)}
        """

        try:
            if verbose:
                logger.debug("[VERBOSE] Requesting AI summary of findings...")
            # Use Abstra's prompt function instead of direct OpenAI API
            summary_response = prompt(summary_prompt, response_format)  # type: ignore
            if verbose:
                logger.debug(f"[VERBOSE] Summary response: {summary_response}")
            return summary_response
        except Exception as e:
            logger.warning("Error summarizing findings", exc_info=e)
            if verbose:
                logger.debug(f"[VERBOSE] Error during summarization: {e}")

        observations_count = len(
            [step.observation for step in state.steps if step.observation]
        )
        return {
            "error": "Maximum steps reached. Task may be incomplete.",
            "observations_count": observations_count,
        }

    finally:
        if verbose:
            logger.debug("[VERBOSE] Closing WebDriver and finishing browser session")
        driver.quit()



from abc import ABC, abstractmethod
from typing import Dict, Optional

from abstra_internals.cloud_api.http_client import HTTPClient
from abstra_internals.contracts_generated import CommonUser, CommonUserRoles
from abstra_internals.credentials import resolve_headers


class UsersRepository(ABC):
    @abstractmethod
    def get_user(self, email: str) -> Optional[CommonUser]:
        pass

    @abstractmethod
    def insert_user(self, email: str) -> bool:
        pass


class TestUsersRepository(UsersRepository):
    ## prevent pytest to inspect this class
    __test__ = False
    data: Dict[str, CommonUserRoles]

    def __init__(self, data: Dict[str, CommonUserRoles] = {}) -> None:
        super().__init__()
        self.data = data

    def get_user(self, email: str) -> Optional[CommonUser]:
        if email in self.data:
            return CommonUser(
                id=email,
                email=email,
                roles=self.data[email],
                created_at="2021-09-01T00:00:00Z",
                project_id="1",
            )

        return None

    def insert_user(self, email: str) -> bool:
        if email in self.data:
            return False
        self.data[email] = []
        return True


class LocalUsersRepository(UsersRepository):
    @property
    def headers(self):
        return resolve_headers()

    def get_user(self, email: str) -> Optional[CommonUser]:
        return None

    def insert_user(self, email: str) -> bool:
        return True


class ProductionUsersRepository(UsersRepository):
    def __init__(self, client: "HTTPClient") -> None:
        self.client = client

    def get_user(self, email: str) -> Optional[CommonUser]:
        r = self.client.get(
            endpoint="/users",
            params={"email": email},
        )
        if not r.ok:
            return None
        return CommonUser.from_dict(r.json())

    def insert_user(self, email: str) -> bool:
        r = self.client.post(
            endpoint="/users",
            json={"email": email},
        )
        return r.ok



from collections.abc import Iterable, Sized
from dataclasses import MISSING, fields, is_dataclass
from inspect import Parameter, signature
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
)

from pydantic import BaseModel


def get_function_metadata(func: Callable) -> Dict[str, Any]:
    """Extract metadata from a function for JSON schema generation"""
    func_name = func.__name__
    input_schema = get_function_json_schema(func)

    requires_approval = False
    if hasattr(func, "_requires_approval"):
        requires_approval = func._requires_approval  # type: ignore
        func_name += "__req_approval__"

    return {
        "name": func_name,
        "description": func.__doc__.strip() if func.__doc__ else f"Tool: {func_name}",
        "inputSchema": input_schema,
        "requiresApproval": requires_approval,
    }


def get_function_json_schema(func: Callable) -> Dict[str, Any]:
    func_signature = signature(func)
    parameters = list(func_signature.parameters.values())

    properties = {}
    required = []

    for param in parameters:
        param_name = param.name
        properties[param_name] = get_param_json_schema(param)

        if param.default == Parameter.empty:
            required.append(param_name)

    input_schema = {"type": "object", "additionalProperties": True}

    if properties:
        input_schema["properties"] = properties
        input_schema["required"] = required

    return input_schema


def get_param_json_schema(param: Parameter):
    param_type = param.annotation

    if param_type == Parameter.empty:
        # No type annotation - default to string
        param_schema = {"type": "string"}
    else:
        param_schema = type_to_json_schema(param_type)

    return param_schema


def type_to_json_schema(field_type: Any) -> Dict[str, Any]:
    """Convert a Python type to JSON schema format"""

    # Handle Any
    if field_type is Any:
        return {}

    # Handle NoneType
    if field_type is type(None):
        return {"type": "null"}

    # Handle basic types
    if field_type is str:
        return {"type": "string"}
    elif field_type is int:
        return {"type": "integer"}
    elif field_type is float:
        return {"type": "number"}
    elif field_type is bool:
        return {"type": "boolean"}
    elif field_type is dict or field_type is Dict:
        return {"type": "object", "additionalProperties": True}
    elif field_type is list or field_type is List:
        return {"type": "array", "items": {"type": "string"}}
    elif field_type is tuple or field_type is Tuple:
        return {"type": "array"}

    # Handle typing module types
    origin = get_origin(field_type)
    args = get_args(field_type)

    if origin is Literal:
        return {"enum": list(args)}

    # Handle tuple and Tuple as fixed-length arrays
    if origin in (tuple, Tuple):
        if args:
            item_schemas = [type_to_json_schema(arg) for arg in args]
            return {
                "type": "array",
                "items": item_schemas,
                "minItems": len(item_schemas),
                "maxItems": len(item_schemas),
            }
        else:
            # Generic tuple
            return {"type": "array"}

    # Handle list/array generics
    if origin is list or origin is List:
        if args:
            item_schema = type_to_json_schema(args[0])
            return {"type": "array", "items": item_schema}
        else:
            return {"type": "array", "items": {"type": "string"}}

    # Handle dict/object generics
    if origin is dict or origin is Dict:
        if len(args) >= 2:
            value_schema = type_to_json_schema(args[1])
            return {"type": "object", "additionalProperties": value_schema}
        else:
            return {"type": "object", "additionalProperties": True}

    elif origin is Union:
        # Handle empty Union
        if not args:
            return {}
        # Handle Optional (Union[T, None]) and other unions
        non_none_args = [arg for arg in args if arg is not type(None)]
        if len(non_none_args) == 0:
            # Only NoneType
            return {"type": "null"}
        elif len(non_none_args) == 1:
            # This is Optional[T]
            schema = type_to_json_schema(non_none_args[0])
            # JSON Schema doesn't have a direct "optional" concept, but we can indicate it's nullable
            if isinstance(schema, dict):
                schema = schema.copy()
                schema["nullable"] = True
            return schema
        else:
            # Multiple types union - use anyOf
            return {"anyOf": [type_to_json_schema(arg) for arg in non_none_args]}

    # Handle custom classes (dataclasses, Pydantic models, etc.)
    elif is_dataclass(field_type):
        return dataclass_to_json_schema(field_type)
    elif hasattr(field_type, "model_json_schema"):
        # Pydantic model
        return pydantic_to_json_schema(field_type)

    # Default fallback
    return {"type": "string", "description": f"Unsupported type: {field_type}"}


def pydantic_to_json_schema(model: Type[BaseModel]) -> Dict[str, Any]:
    """Convert a Pydantic model to JSON schema format for MCP"""

    schema = model.model_json_schema()

    # Convert Pydantic schema to MCP-compatible format
    mcp_schema = {
        "type": "object",
        "properties": schema.get("properties", {}),
        "required": schema.get("required", []),
        "additionalProperties": True,
    }

    return mcp_schema


def dataclass_to_json_schema(dataclass_type: Any) -> Dict[str, Any]:
    """Convert a dataclass to JSON schema format for MCP"""

    if not is_dataclass(dataclass_type):
        raise ValueError("Type must be a dataclass")

    properties = {}
    required = []

    for field in fields(dataclass_type):
        field_type = field.type

        # Use the enhanced type conversion
        properties[field.name] = type_to_json_schema(field_type)

        # Check if field is required (no default value)
        if field.default is MISSING and field.default_factory is MISSING:
            required.append(field.name)

    return {
        "type": "object",
        "properties": properties,
        "required": required,
        "additionalProperties": True,
    }


def validate_type(value: object, schema: dict) -> bool:
    # Handle anyOf at the top level
    if "anyOf" in schema:
        return any(validate_type(value, s) for s in schema["anyOf"])

    t = schema.get("type")
    if t == "string":
        return isinstance(value, str)
    elif t == "integer":
        return isinstance(value, int) and not isinstance(value, bool)
    elif t == "number":
        return isinstance(value, (int, float)) and not isinstance(value, bool)
    elif t == "boolean":
        return isinstance(value, bool)
    elif t == "null":
        return value is None
    elif t == "array":
        items = schema.get("items", {"type": "string"})
        # Ensure value is Sized and Iterable before using len() or iterating
        if not (isinstance(value, Sized) and isinstance(value, Iterable)):
            return False
        # Handle tuple validation (items is a list of schemas)
        if isinstance(items, list):
            if len(value) != len(items):
                return False
            for v, item_schema in zip(value, items):
                if not validate_type(v, item_schema):
                    return False
            # Optionally check minItems/maxItems
            if "minItems" in schema and len(value) < schema["minItems"]:
                return False
            if "maxItems" in schema and len(value) > schema["maxItems"]:
                return False
            return True
        else:
            # Regular list validation
            if "minItems" in schema and len(value) < schema["minItems"]:
                return False
            if "maxItems" in schema and len(value) > schema["maxItems"]:
                return False
            for v in value:
                if not validate_type(v, items):
                    return False
            return True
    elif t == "object":
        if not isinstance(value, dict):
            return False
        properties = schema.get("properties", {})
        required = schema.get("required", [])
        additional = schema.get("additionalProperties", True)
        # Check required fields
        for req in required:
            if req not in value:
                return False
        # Check properties
        for prop, prop_schema in properties.items():
            if prop in value and not validate_type(value[prop], prop_schema):
                return False
        # Check additionalProperties
        if not additional:
            # No additional properties allowed
            for k in value:
                if k not in properties:
                    return False
        elif isinstance(additional, dict):
            # additionalProperties is a schema
            for k in value:
                if k not in properties:
                    if not validate_type(value[k], additional):
                        return False
        return True
    # fallback: accept anything
    return True



import flask

from abstra_internals.contracts_generated import AbstraLibApiEditorLintersFixResponse
from abstra_internals.controllers.main import MainController
from abstra_internals.usage import editor_usage


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_linters", __name__)

    # 1s pooling in this route
    @bp.get("/check")
    def _check_linters():
        checks = controller.linter_repository.find_issues_in_codebase()
        return [check.to_dict() for check in checks]

    @bp.post("/fix/<rule_name>/<fix_name>")
    @editor_usage
    def _fix_linter(rule_name: str, fix_name: str):
        controller.linter_repository.fix_issue_in_codebase(rule_name, fix_name)
        return AbstraLibApiEditorLintersFixResponse(success=True).to_dict()

    @bp.post("/fix-all")
    @editor_usage
    def _fix_all_linters():
        controller.linter_repository.fix_all_linters()
        return AbstraLibApiEditorLintersFixResponse(success=True).to_dict()

    return bp



from typing import Optional

from abstra_internals.email_templates import task_waiting_template
from abstra_internals.entities.execution import Execution
from abstra_internals.entities.execution_context import ScriptContext
from abstra_internals.repositories.factory import Repositories
from abstra_internals.repositories.project.project import (
    ComponentStage,
    FormStage,
    ScriptStage,
    Stage,
)
from abstra_internals.repositories.tasks import TaskDTO, TaskPayload


class TaskExecutor:
    def __init__(self, repos: Repositories) -> None:
        self.repos = repos
        self.project = self.repos.project.load(include_disabled_stages=False)

    def send_task(
        self,
        type: str,
        current_stage: Stage,
        payload: TaskPayload,
        execution: Optional[Execution] = None,
        show_warning: bool = True,
    ) -> None:
        project = self.repos.project.load(include_disabled_stages=False)

        next_stages = []
        module_name = project.find_module_containing_stage(current_stage.id)
        if module_name is not None:
            if project.is_stage_module_output(current_stage.id):
                parent_component_stage = project.get_component_by_module_name(
                    module_name
                )
                if parent_component_stage is not None:
                    next_stages = [
                        project.get_stage_raises(t.target_id)
                        for t in parent_component_stage.workflow_transitions
                        if t.matches(type)
                    ]
            else:
                next_stages = [
                    project.get_stage_raises(t.target_id)
                    for t in current_stage.workflow_transitions
                    if t.matches(type)
                ]
        else:
            next_stages = [
                project.get_stage_raises(t.target_id)
                for t in current_stage.workflow_transitions
                if t.matches(type)
            ]

        if len(next_stages) == 0 and show_warning:
            print(
                f"[WARNING] No transitions found for task type {type} in stage {current_stage.id}"
            )
            return

        for stage in next_stages:
            if isinstance(stage, ComponentStage):
                module = stage.package_name
                if module is not None:
                    input_stages = self.project.get_inputs_for_module(module)
                    next_stages.remove(stage)
                    next_stages.extend(input_stages)

        for stage in next_stages:
            task = self.repos.tasks.send_task(
                type=type,
                payload=payload,
                source_stage_id=current_stage.id,
                target_stage_id=stage.id,
                execution_id=execution.id if execution else None,
            )
            self._send_waiting_thread_notification(task)
            if execution:
                execution.context.sent_tasks.append(task.id)
            if isinstance(stage, ScriptStage):
                self.repos.producer.enqueue_fire_and_forget(
                    context=ScriptContext(task_id=task.id),
                    stage_id=stage.id,
                )

    def _send_waiting_thread_notification(self, task: TaskDTO):
        stage = self.project.get_stage(task.target_stage_id)
        if not stage:
            raise Exception(f"Stage {task.target_stage_id} not found")

        if not (isinstance(stage, FormStage) and stage.notification_trigger.enabled):
            return

        recipient_emails = stage.notification_trigger.get_recipients(task.payload)
        if not recipient_emails:
            return

        self.repos.email.send(
            task_waiting_template.generate_email(
                recipient_emails=recipient_emails, form=stage
            )
        )



import threading
import time
from pathlib import Path, PurePath
from typing import Callable, Dict, List, Literal, Optional, Union

from watchdog.events import (
    FileCreatedEvent,
    FileDeletedEvent,
    FileModifiedEvent,
    FileMovedEvent,
    FileSystemEvent,
    FileSystemEventHandler,
)
from watchdog.observers import Observer

from abstra_internals.settings import Settings
from abstra_internals.utils.crdt import CRDTManager
from abstra_internals.utils.file import safe_read_file, safe_write_file

IGNORED_PATHS = [
    ".abstra/",
    ".venv",
    "__pycache__",
    ".git/index.lock",
    ".git/HEAD.lock",
    ".git/config.lock",
    ".git/refs/",
    ".git/objects/",
    ".git/hooks/",
    ".git/logs/",
    ".git/COMMIT_EDITMSG",
    ".git/MERGE_HEAD",
    ".git/FETCH_HEAD",
    ".git/ORIG_HEAD",
]
FSEventType = Literal["changed", "created", "deleted", "moved"]
Handler = Callable[[Path, FSEventType, Optional[str]], None]

crdt_managers: Dict[str, CRDTManager] = {}


class FileWatcher(FileSystemEventHandler):
    def __init__(self, handlers: List[Handler]):
        super().__init__()
        self._debounce_timers: dict[str, threading.Timer] = {}
        self._modules_folder_timer: Optional[threading.Timer] = None
        self.handlers: List[Handler] = handlers

    def start(self):
        observer = Observer()
        observer.schedule(self, path=str(Settings.root_path), recursive=True)
        observer.start()
        self._observer = observer

    def dispatch(self, event: FileSystemEvent):
        filepath = Path(event.src_path).absolute()
        filepath_str = str(filepath)

        if self.should_ignore_path(filepath):
            return

        content = None
        if isinstance(event, FileCreatedEvent):
            event_type = "created"
        elif isinstance(event, FileDeletedEvent):
            event_type = "deleted"
        elif isinstance(event, FileMovedEvent):
            event_type = "moved"
        elif isinstance(event, FileModifiedEvent):
            event_type = "changed"
            content = safe_read_file(filepath, 2.0)
            if not content:
                return

            if filepath_str not in crdt_managers or crdt_managers[filepath_str] is None:
                crdt_managers[filepath_str] = CRDTManager(file_path=filepath)

            old_content = crdt_managers[filepath_str].get_content()
            if old_content != content:
                crdt_managers[filepath_str].apply_operations_from_diff(content)
                new_content = crdt_managers[filepath_str].get_content()
                if content != new_content:
                    safe_write_file(filepath, new_content, 2.0)

        else:
            return

        def execute_handlers():
            time.sleep(0.01)
            threads = []
            for handler in self.handlers:
                thread = threading.Thread(
                    target=handler, args=(filepath, event_type, content)
                )
                thread.start()
                threads.append(thread)

            for thread in threads:
                thread.join()

        if "/modules/" in filepath_str or filepath_str.endswith("/modules"):
            if self._modules_folder_timer is not None:
                self._modules_folder_timer.cancel()

            self._modules_folder_timer = threading.Timer(
                interval=2.0, function=execute_handlers
            )
            self._modules_folder_timer.start()
        else:
            if filepath_str in self._debounce_timers:
                self._debounce_timers[filepath_str].cancel()

            self._debounce_timers[filepath_str] = threading.Timer(
                interval=0.5, function=execute_handlers
            )
            self._debounce_timers[filepath_str].start()

    def should_ignore_path(self, path: Union[Path, PurePath]) -> bool:
        path_str = str(path).replace("\\", "/")

        for ignored_pattern in IGNORED_PATHS:
            normalized_pattern = ignored_pattern.replace("\\", "/")

            if "/" in normalized_pattern:
                if normalized_pattern in path_str:
                    return True

                if path_str.endswith(normalized_pattern):
                    return True
            else:
                if (
                    f"/{normalized_pattern}/" in path_str
                    or path_str.endswith(f"/{normalized_pattern}")
                    or path_str.startswith(f"{normalized_pattern}/")
                    or path_str == normalized_pattern
                ):
                    return True

        if "/.git/" in path_str and path_str.endswith((".lock", ".tmp")):
            return True

        return False



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from abstra_internals.cloud_api.http_client import HTTPClient


@dataclass
class WorkerCapacity:
    current_workers: int
    max_workers: int
    is_at_capacity: bool

    @staticmethod
    def from_dict(data: dict) -> "WorkerCapacity":
        return WorkerCapacity(
            current_workers=data.get("currentWorkers", 0),
            max_workers=data.get("maxWorkers", 0),
            is_at_capacity=data.get("isAtCapacity", False),
        )


class InfraRepository(ABC):
    @abstractmethod
    def get_worker_capacity(self) -> Optional[WorkerCapacity]:
        raise NotImplementedError()


class LocalInfraRepository(InfraRepository):
    def get_worker_capacity(self) -> Optional[WorkerCapacity]:
        return None


class ProductionInfraRepository(InfraRepository):
    def __init__(self, client: HTTPClient):
        self.client = client

    def get_worker_capacity(self) -> Optional[WorkerCapacity]:
        try:
            res = self.client.get("/infra/worker-capacity")
            res.raise_for_status()
            return WorkerCapacity.from_dict(res.json())
        except Exception:
            return None



from abstra_internals.services.sql_storage import SqlStorage
from abstra_internals.utils.serializable import Serializable
from tests.fixtures import BaseTest


class MockModel(Serializable):
    name: str
    age: int


class ModelWithReservedWords(Serializable):
    """Model with SQL reserved words as field names to test escaping."""

    update: str
    select: str
    where: str


class TestFileManager(BaseTest):
    def setUp(self) -> None:
        super().setUp()
        mp_context = self.repositories.mp_context.get_context()
        self.manager = SqlStorage[MockModel](
            mp_context, directory="test", model=MockModel
        )

    def test_save_and_load(self):
        test_data = MockModel(name="John", age=30)
        self.manager.save("test_id", test_data)

        loaded_data = self.manager.load("test_id")
        assert loaded_data is not None
        assert loaded_data.name == "John"
        assert loaded_data.age == 30

    def test_load_all(self):
        test_data1 = MockModel(name="John", age=30)
        test_data2 = MockModel(name="Jane", age=25)

        self.manager.save("test_id_1", test_data1)
        self.manager.save("test_id_2", test_data2)

        loaded_data = self.manager.load_all()
        assert len(loaded_data) == 2

    def test_delete(self):
        test_data = MockModel(name="John", age=30)
        self.manager.save("test_id", test_data)

        self.manager.delete("test_id")
        loaded_data = self.manager.load("test_id")
        assert loaded_data is None

    def test_clear(self):
        test_data1 = MockModel(name="John", age=30)
        test_data2 = MockModel(name="Jane", age=25)

        self.manager.save("test_id_1", test_data1)
        self.manager.save("test_id_2", test_data2)

        self.manager.clear()
        loaded_data = self.manager.load_all()
        assert len(loaded_data) == 0

    def test_load_nonexistent(self):
        loaded_data = self.manager.load("nonexistent_id")
        assert loaded_data is None

    def test_save_invalid_data(self):
        # This test needs to be adapted for SQL storage
        # Instead of writing invalid data directly to a file,
        # we test that loading a model with invalid data fails gracefully
        # by trying to create a model with invalid data
        try:
            invalid_model = MockModel(name="John", age="thirty")  # type: ignore
            # If this succeeds, pydantic coerced the string to int
            # So we save it and load it back
            self.manager.save("test_id", invalid_model)
            loaded_data = self.manager.load("test_id")
            # The data should load successfully since pydantic accepted it
            assert loaded_data is not None
        except Exception:
            # If pydantic rejects it, that's also acceptable
            pass

    def test_reserved_sql_keywords(self):
        """Test that SQL reserved words as field names are properly escaped."""
        mp_context = self.repositories.mp_context.get_context()
        manager = SqlStorage[ModelWithReservedWords](
            mp_context, directory="test_reserved", model=ModelWithReservedWords
        )

        # Test save with reserved keywords
        test_data = ModelWithReservedWords(
            update="update_value", select="select_value", where="where_value"
        )
        manager.save("reserved_id", test_data)

        # Test load
        loaded_data = manager.load("reserved_id")
        assert loaded_data is not None
        assert loaded_data.update == "update_value"
        assert loaded_data.select == "select_value"
        assert loaded_data.where == "where_value"

        # Test update
        test_data.update = "new_update_value"
        manager.save("reserved_id", test_data)

        reloaded_data = manager.load("reserved_id")
        assert reloaded_data is not None
        assert reloaded_data.update == "new_update_value"



import unittest
from unittest.mock import MagicMock

from abstra_internals.server.apps.cloud import get_cloud_app


class TestSecurityPlayerStop(unittest.TestCase):
    def setUp(self):
        # Mock the controller and repositories
        self.mock_controller = MagicMock()
        # Ensure repositories structure exists
        self.mock_controller.repositories = MagicMock()
        self.mock_controller.repositories.project = MagicMock()

        # Create the Player App (Cloud App)
        self.app = get_cloud_app(self.mock_controller)
        self.client = self.app.test_client()

    def test_stop_route_not_exposed_on_player(self):
        # The stop route is defined as POST /api/executions/<id>/stop in the Editor
        # or /executions/<id>/stop depending on the blueprint registration.
        # In the Editor Blueprint:
        #   bp.register_blueprint(executions_bp, url_prefix="/executions")
        #   bp.register_blueprint(api_bp, url_prefix="/api") -> _get_api_bp -> executions
        # So it's likely /api/executions/<id>/stop or /executions/<id>/stop in the Editor context.

        # We try to access it on the Player App.

        execution_id = "exec-test-security"

        # Try various potential paths just to be sure
        paths_to_test = [
            f"/executions/{execution_id}/stop",
            f"/api/executions/{execution_id}/stop",
            f"/_editor/executions/{execution_id}/stop",
        ]

        for path in paths_to_test:
            response = self.client.post(path)
            # Should be 404 Not Found (if no route matches) or 405 Method Not Allowed
            # (if it matches the GET-only SPA catch-all).
            # Crucially, it must NOT be 200 (OK) or 202 (Accepted).
            self.assertIn(
                response.status_code,
                [404, 405],
                f"Path {path} returned {response.status_code}",
            )


if __name__ == "__main__":
    unittest.main()



from dataclasses import dataclass

from abstra_internals.environment import (
    ABSTRA_EXECUTOR_ACQUIRE_TIMEOUT,
    ABSTRA_EXECUTOR_MAX_EXECUTIONS,
    ABSTRA_EXECUTOR_POOL_SIZE,
    ABSTRA_EXECUTOR_WARMUP_PARALLELISM,
    ABSTRA_EXECUTOR_WARMUP_TIMEOUT,
    ABSTRA_MIN_FORMS_EXECUTORS,
    PROCESS_TIMEOUT_SECONDS,
)


@dataclass(frozen=True)
class ExecutorConfig:
    total_executors: int
    min_form_executors: int
    max_executions_per_executor: int
    warmup_timeout_seconds: float
    execution_timeout_seconds: float
    warmup_parallelism: int
    acquire_timeout_seconds: float

    @classmethod
    def from_environment(cls) -> "ExecutorConfig":
        total_executors = ABSTRA_EXECUTOR_POOL_SIZE
        min_form_executors = ABSTRA_MIN_FORMS_EXECUTORS
        max_executions = ABSTRA_EXECUTOR_MAX_EXECUTIONS
        warmup_timeout = ABSTRA_EXECUTOR_WARMUP_TIMEOUT
        warmup_parallelism = ABSTRA_EXECUTOR_WARMUP_PARALLELISM
        acquire_timeout = ABSTRA_EXECUTOR_ACQUIRE_TIMEOUT

        return cls(
            total_executors=total_executors,
            min_form_executors=min_form_executors,
            max_executions_per_executor=max_executions,
            warmup_timeout_seconds=warmup_timeout,
            execution_timeout_seconds=PROCESS_TIMEOUT_SECONDS,
            warmup_parallelism=warmup_parallelism,
            acquire_timeout_seconds=acquire_timeout,
        )

    def validate(self) -> None:
        if self.total_executors < 1:
            raise ValueError("total_executors must be at least 1")

        if self.min_form_executors < 0:
            raise ValueError("min_form_executors cannot be negative")

        if self.min_form_executors > self.total_executors:
            raise ValueError("min_form_executors cannot exceed total_executors")

        if self.max_executions_per_executor < 1:
            raise ValueError("max_executions_per_executor must be at least 1")

        if self.warmup_timeout_seconds <= 0:
            raise ValueError("warmup_timeout_seconds must be positive")

        if self.execution_timeout_seconds <= 0:
            raise ValueError("execution_timeout_seconds must be positive")

        if self.warmup_parallelism < 1:
            raise ValueError("warmup_parallelism must be at least 1")

        if self.acquire_timeout_seconds <= 0:
            raise ValueError("acquire_timeout_seconds must be positive")



from typing import Dict

import abstra_internals.interface.contract as contract
from abstra_internals.controllers.execution.connection_protocol import (
    ConnectionProtocol,
)
from abstra_internals.controllers.execution.execution_client import ExecutionClient
from abstra_internals.entities.execution_context import HookContext, Request, Response
from abstra_internals.utils import serialize


class HookClient(ExecutionClient):
    context: HookContext
    conn: ConnectionProtocol
    response: Response
    production_mode: bool

    def __init__(
        self, context: HookContext, conn: ConnectionProtocol, production_mode: bool
    ) -> None:
        self.context = context
        self.conn = conn
        self.response = Response(headers={}, status=200, body="")
        self.production_mode = production_mode

    def handle_failure(self, e: Exception) -> None:
        self.set_response(500, "An exception occurred during execution.", {})
        try:
            self.conn.send(self.response)
            close_dto = contract.CloseDTO(exit_status="EXCEPTION", exception=e)
            self._send(contract.ExecutionEndedMessage(close_dto, self.production_mode))
        except (BrokenPipeError, EOFError):
            pass

    def handle_success(self) -> None:
        try:
            self.conn.send(self.response)
            close_dto = contract.CloseDTO(exit_status="SUCCESS")
            self._send(contract.ExecutionEndedMessage(close_dto, self.production_mode))
        except (BrokenPipeError, EOFError):
            pass

    def set_response(self, status: int, body: str, headers: Dict[str, str]) -> None:
        self.response = Response(
            status=status,
            body=body,
            headers=headers,
        )

    def get_request(self) -> Request:
        return self.context.request

    def handle_start(self, execution_id: str):
        self._send(contract.ExecutionStartedMessage(execution_id))

    def _send(self, msg: contract.Message) -> None:
        str_data = serialize(msg.to_json())
        try:
            self.conn.send(str_data)
        except (EOFError, BrokenPipeError):
            pass



import shutil
import time
import unittest
from dataclasses import dataclass
from pathlib import Path
from uuid import uuid4

from abstra_internals.entities.execution import Execution
from abstra_internals.entities.execution_context import (
    FormContext,
    FormExecutionMock,
    Request,
)
from abstra_internals.repositories.execution import ExecutionFilter
from abstra_internals.repositories.factory import build_editor_repositories
from abstra_internals.repositories.multiprocessing import SpawnContextReposity
from abstra_internals.settings import Settings


@dataclass
class FakeStage:
    """Fake stage to simulate what is passed to the child process"""

    id: str
    title: str
    file: str


def child_process_main(stage: FakeStage, root_path: str):
    """Function executed in the child process (worker)."""
    Settings.set_root_path(root_path)

    # Try to create repositories (this used to cause the error before the fix)
    repositories = build_editor_repositories()

    # Try to create an execution (write to SqlStorage)
    execution = Execution.create(
        id=uuid4().__str__(),
        stage_id=stage.id,
        context=FormContext(
            request=Request(
                query_params={},
                headers={},
                method="GET",
                body="{}",
            ),
            mock_execution=FormExecutionMock(test_answers=[]),
        ),
        worker_id="mock-worker-id",
    )

    repositories.execution.create(execution)

    # Try to update the execution
    execution.set_status("finished")
    repositories.execution.update(execution)

    # Try to read executions
    filter = ExecutionFilter(offset=0, limit=10)
    all_executions = repositories.execution.list(filter)

    return len(all_executions.executions) > 0


class TestSemLockFix(unittest.TestCase):
    """Test script to verify that the SemLock error has been resolved.

    Reproduces the scenario that caused the error:
    1. PARENT process creates SqlStorage (as web-editor does)
    2. Tries to spawn CHILD process passing objects with SqlStorage
    3. CHILD process tries to use SqlStorage

    Before the fix: error "SemLock created in a fork context being shared with spawn context"
    After the fix: should work normally
    """

    def setUp(self):
        self.test_dir = Path(__file__).parent / ".abstra_test"
        self.test_dir.mkdir(exist_ok=True)
        Settings.set_root_path(str(self.test_dir))

    def tearDown(self):
        if self.test_dir.exists():
            shutil.rmtree(self.test_dir)

    def test_semlock_scenario(self):
        """Tests the complete scenario that caused the SemLock error"""
        # PARENT PROCESS: creates repositories (as web-editor does in editor.py:92)
        build_editor_repositories()

        # Create a fake stage to pass to the child process
        stage = FakeStage(id="test-form-123", title="Test Form", file="test.py")

        # SPAWN CHILD PROCESS: uses spawn context (as consumer.py:84 does)
        mp_context = SpawnContextReposity().get_context()

        process = mp_context.Process(
            target=child_process_main,
            args=(stage, str(self.test_dir)),
            name="TestWorker",
        )

        process.start()
        process.join(timeout=30)

        # Ensure process actually finished
        if process.is_alive():
            process.terminate()
            process.join(timeout=2)

        self.assertFalse(process.is_alive(), "Process should not be alive")
        self.assertEqual(
            process.exitcode,
            0,
            "Child process should have completed without errors (no SemLock error)",
        )

    def test_concurrent_workers(self):
        """Tests scenario with multiple simultaneous workers (like the real system)"""
        build_editor_repositories()

        mp_context = SpawnContextReposity().get_context()
        processes = []

        for i in range(2):
            stage = FakeStage(
                id=f"test-form-{i}", title=f"Test Form {i}", file=f"test{i}.py"
            )

            process = mp_context.Process(
                target=child_process_main,
                args=(stage, str(self.test_dir)),
                name=f"TestWorker-{i}",
            )

            processes.append(process)
            process.start()
            time.sleep(0.1)  # Stagger process starts to reduce contention

        # Wait for all processes
        for process in processes:
            process.join(timeout=30)
            # Ensure process actually finished
            if process.is_alive():
                process.terminate()
                process.join(timeout=2)

        # Verify that all completed successfully
        for i, process in enumerate(processes):
            self.assertFalse(process.is_alive(), f"Worker {i} should not be alive")
            self.assertEqual(
                process.exitcode, 0, f"Worker {i} should have completed successfully"
            )



import time
import unittest
from multiprocessing import Queue
from unittest.mock import MagicMock

from abstra_internals.controllers.execution.consumer import ConsumerController
from abstra_internals.repositories.consumer import (
    Consumer,
    ControlMessage,
    ControlQueueMessage,
)
from abstra_internals.settings import SettingsController


class MockExecutorPool:
    def __init__(self):
        self.killed_ids = []

    def kill_execution(self, execution_id):
        # Simulate check
        if execution_id == "explode":
            raise Exception("Unexpected error in pool")
        self.killed_ids.append(execution_id)

    def start(self):
        pass

    def shutdown(self):
        pass

    def can_start_loop(self):
        return True


class FaultyControlConsumer(Consumer):
    def __init__(self, queue: Queue):
        self.queue = queue
        self.running = True
        self.acks = []
        self.nacks = []

    def threadsafe_ack(self, msg):
        self.acks.append(msg)

    def threadsafe_nack(self, msg):
        self.nacks.append(msg)

    def stop_iter(self):
        self.running = False

    def iter(self):
        while self.running:
            try:
                # Using a slightly longer timeout to prevent tight loops in tests
                msg = self.queue.get(timeout=0.1)
                yield msg
            except Exception:
                continue


class TestWebEditorKillRobustness(unittest.TestCase):
    def setUp(self):
        SettingsController.set_root_path("/tmp")
        SettingsController.set_server_port(3000)
        self.control_queue = Queue()
        self.consumer_mock = FaultyControlConsumer(self.control_queue)

        self.mock_main_controller = MagicMock()
        self.mock_main_controller.repositories.mp_context.get_context.return_value = (
            MagicMock()
        )
        self.mock_main_controller.repositories.producer = MagicMock()

        # Mock the regular consumer to just sit idle
        self.regular_consumer = MagicMock()
        self.regular_consumer.iter.return_value = []

        self.controller = ConsumerController(
            self.mock_main_controller,
            self.regular_consumer,
            control_consumer=self.consumer_mock,
        )
        self.pool_mock = MockExecutorPool()
        self.controller.executor_pool = self.pool_mock

    def tearDown(self):
        self.consumer_mock.stop_iter()
        self.controller.shutdown()

    def test_malformed_message_does_not_crash_loop(self):
        # 1. Start the loop
        self.controller.start_loop()

        # 2. Send a "Poison Message" (Not a ControlQueueMessage, just random string/object)
        # The real RabbitMQ consumer might yield deserialization errors, but if logic fails
        # inside the loop, it might receive something unexpected or raise.
        # Here we verify the loop handles unexpected types gracefully.
        poison_msg = "I am not a message"
        self.control_queue.put(poison_msg)

        # 3. Send a VALID message immediately after
        valid_id = "exec-valid-1"
        control_msg = ControlMessage(type="stop", payload={"execution_id": valid_id})
        queue_msg = ControlQueueMessage(message=control_msg, delivery_tag=1)
        self.control_queue.put(queue_msg)

        # 4. Wait for processing
        time.sleep(1.0)  # Give thread time to process both

        # 5. Assertions
        # The loop should have survived the poison message and processed the valid one
        self.assertIn(valid_id, self.pool_mock.killed_ids)
        # The valid message should be acked
        self.assertIn(queue_msg, self.consumer_mock.acks)

        # Depending on implementation, the poison message might be acked (to clear) or nacked
        # Our current implementation acks unknown types to clear the queue (ConsumerController._control_loop logic)
        # "Received unknown message type... self.control_consumer.threadsafe_ack(msg)"
        self.assertIn(poison_msg, self.consumer_mock.acks)

    def test_exception_during_processing_keeps_loop_alive(self):
        # Risk: ExecutorPool raises an exception. Does the loop crash?
        self.controller.start_loop()

        # 1. Send message that triggers exception in our MockPool
        control_msg_bad = ControlMessage(
            type="stop", payload={"execution_id": "explode"}
        )
        queue_msg_bad = ControlQueueMessage(message=control_msg_bad, delivery_tag=666)
        self.control_queue.put(queue_msg_bad)

        # 2. Send valid message after
        control_msg_good = ControlMessage(
            type="stop", payload={"execution_id": "exec-good"}
        )
        queue_msg_good = ControlQueueMessage(message=control_msg_good, delivery_tag=777)
        self.control_queue.put(queue_msg_good)

        time.sleep(1.0)

        # 3. Assertions
        # The second message MUST be processed implies the thread is still alive
        self.assertIn("exec-good", self.pool_mock.killed_ids)

        # The bad message should ideally be NACKed if an exception occurred
        # Check ConsumerController logic: "except Exception ... threadsafe_nack(msg)"
        self.assertIn(queue_msg_bad, self.consumer_mock.nacks)
        self.assertIn(queue_msg_good, self.consumer_mock.acks)

    def test_kill_unknown_execution_is_safe(self):
        # This primarily tests the ExecutorPool, but here we verify integration
        self.controller.start_loop()

        # 1. Kill ID that doesn't exist
        control_msg = ControlMessage(type="stop", payload={"execution_id": "ghost-id"})
        queue_msg = ControlQueueMessage(message=control_msg, delivery_tag=999)
        self.control_queue.put(queue_msg)

        time.sleep(0.5)

        # 2. Assert it called the pool (pool handles the logic of "not found")
        self.assertIn("ghost-id", self.pool_mock.killed_ids)
        self.assertIn(queue_msg, self.consumer_mock.acks)


if __name__ == "__main__":
    unittest.main()
